# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Common test-helper code involving Tasks."""

import abc
import datetime as dt
import hashlib
import shlex
import tarfile
from collections.abc import Collection as AbcCollection
from collections.abc import Mapping, Sequence
from io import BytesIO
from pathlib import Path
from shutil import rmtree
from tempfile import mkdtemp
from typing import Any, Protocol, cast, overload, override
from unittest import mock
from unittest.mock import MagicMock

from debusine.artifacts.models import ArtifactCategory, CollectionCategory
from debusine.client.debusine import Debusine
from debusine.client.models import (
    ArtifactResponse,
    FileResponse,
    FilesResponseType,
    RelationType,
)
from debusine.tasks import BaseExternalTask, BaseTask, BaseTaskWithExecutor
from debusine.tasks.executors import ExecutorInterface, UnshareInstance
from debusine.tasks.models import (
    BaseDynamicTaskData,
    BaseDynamicTaskDataWithExecutor,
    BaseTaskData,
    BaseTaskDataWithExecutor,
    LookupMultiple,
    LookupSingle,
)
from debusine.tasks.server import (
    ArtifactInfo,
    CollectionInfo,
    MultipleArtifactInfo,
    TaskDatabaseInterface,
)
from debusine.test.base import TestCase
from debusine.test.test_utils import create_artifact_response
from debusine.utils import extract_generic_type_arguments


class TaskTestProtocol[TaskClass: BaseTask[Any, Any]](Protocol):
    """Attributes that tests using :py:class:`TaskHelperMixin` must provide."""

    SAMPLE_TASK_DATA: dict[str, Any]
    task_class: type[TaskClass]
    task: TaskClass


class MockTaskDatabase(TaskDatabaseInterface):
    """
    Task database implementation that resolves fields by direct injection.

    This is sufficient for testing task implementations.

    Lookups raise NotImplementedError.
    """

    def __init__(self, **field_values: Any) -> None:
        """Construct a :py:class:`MockTaskDatabase`."""
        self.field_values: dict[str, Any] = {**field_values}

    @overload
    def lookup_single_artifact(
        self,
        lookup: LookupSingle,
        default_category: CollectionCategory | None = None,
    ) -> ArtifactInfo: ...

    @overload
    def lookup_single_artifact(
        self, lookup: None, default_category: CollectionCategory | None = None
    ) -> None: ...

    @override
    def lookup_single_artifact(
        self,
        lookup: LookupSingle | None,
        default_category: CollectionCategory | None = None,
    ) -> ArtifactInfo | None:
        raise NotImplementedError("lookup_single_artifact")

    @override
    def lookup_multiple_artifacts(
        self,
        lookup: LookupMultiple | None,
        default_category: CollectionCategory | None = None,
    ) -> MultipleArtifactInfo:
        raise NotImplementedError("lookup_multiple_artifact")

    @override
    def find_related_artifacts(
        self,
        artifact_ids: AbcCollection[int],
        target_category: ArtifactCategory,
        relation_type: RelationType = RelationType.RELATES_TO,
    ) -> MultipleArtifactInfo:
        raise NotImplementedError("find_related_artifacts")

    @overload
    def lookup_single_collection(
        self,
        lookup: LookupSingle,
        default_category: CollectionCategory | None = None,
    ) -> CollectionInfo: ...

    @overload
    def lookup_single_collection(
        self, lookup: None, default_category: CollectionCategory | None = None
    ) -> None: ...

    @override
    def lookup_single_collection(
        self,
        lookup: LookupSingle | None,
        default_category: CollectionCategory | None = None,
    ) -> CollectionInfo | None:
        raise NotImplementedError("lookup_single_collection")

    @override
    def get_server_setting(self, setting: str) -> str:
        raise NotImplementedError("get_server_setting")

    @override
    def export_suite_signing_keys(self, suite_id: int) -> str:
        raise NotImplementedError("export_suite_signing_keys")

    @override
    def resolve_inputs(self, task: "BaseTask[Any, Any]") -> None:
        for name, inp in task.inputs.items():
            assert name in self.field_values, (
                f"Value for field {task.__class__.__name__}.{name}"
                " was not provided"
            )
            task.__dict__[name] = self.field_values[name]


class FakeTaskDatabaseBase(MockTaskDatabase):
    """
    Fake implementation of database interaction in worker tests.

    Fields not provided for injection are looked up normally.

    This is useful to test more complex dependencies from
    TaskDatabaseInterface, such as input field resolution.
    """

    def __init__(
        self,
        *,
        single_lookups: (
            Mapping[
                tuple[LookupSingle, CollectionCategory | None],
                ArtifactInfo | None,
            ]
            | None
        ) = None,
        multiple_lookups: (
            Mapping[
                tuple[LookupMultiple, CollectionCategory | None],
                list[ArtifactInfo],
            ]
            | None
        ) = None,
        relations: (
            Mapping[
                tuple[Sequence[int], ArtifactCategory, RelationType],
                list[ArtifactInfo],
            ]
            | None
        ) = None,
        single_collection_lookups: (
            Mapping[
                tuple[LookupSingle, CollectionCategory | None],
                CollectionInfo | None,
            ]
            | None
        ) = None,
        settings: dict[str, str] | None = None,
        suite_signing_keys: dict[int, str] | None = None,
        **field_values: Any,
    ) -> None:
        """Construct a :py:class:`FakeTaskDatabase`."""
        super().__init__(**field_values)
        self.single_lookups = dict(single_lookups or {})
        self.multiple_lookups = dict(multiple_lookups or {})
        self.relations = dict(relations or {})
        self.single_collection_lookups = dict(single_collection_lookups or {})
        self.settings = dict(settings or {})
        self.suite_signing_keys = dict(suite_signing_keys or {})

    @overload
    def lookup_single_artifact(
        self,
        lookup: LookupSingle,
        default_category: CollectionCategory | None = None,
    ) -> ArtifactInfo: ...

    @overload
    def lookup_single_artifact(
        self, lookup: None, default_category: CollectionCategory | None = None
    ) -> None: ...

    @override
    def lookup_single_artifact(
        self,
        lookup: LookupSingle | None,
        default_category: CollectionCategory | None = None,
    ) -> ArtifactInfo | None:
        return (
            None
            if lookup is None
            else self.single_lookups[(lookup, default_category)]
        )

    @override
    def lookup_multiple_artifacts(
        self,
        lookup: LookupMultiple | None,
        default_category: CollectionCategory | None = None,
    ) -> MultipleArtifactInfo:
        # if not lookup:
        #     return MultipleArtifactInfo()
        # return MultipleArtifactInfo(
        #     self.multiple_lookups[(lookup, default_category)]
        # )
        raise NotImplementedError("this method is not currently used by tests")

    @override
    def find_related_artifacts(
        self,
        artifact_ids: AbcCollection[int],
        target_category: ArtifactCategory,
        relation_type: RelationType = RelationType.RELATES_TO,
    ) -> MultipleArtifactInfo:
        return MultipleArtifactInfo(
            self.relations[
                (tuple(artifact_ids), target_category, relation_type)
            ]
        )

    @overload
    def lookup_single_collection(
        self,
        lookup: LookupSingle,
        default_category: CollectionCategory | None = None,
    ) -> CollectionInfo: ...

    @overload
    def lookup_single_collection(
        self, lookup: None, default_category: CollectionCategory | None = None
    ) -> None: ...

    def lookup_single_collection(
        self,
        lookup: LookupSingle | None,
        default_category: CollectionCategory | None = None,
    ) -> CollectionInfo | None:
        """Pretend to look up a single collection."""
        return (
            None
            if lookup is None
            else self.single_collection_lookups[(lookup, default_category)]
        )

    def get_server_setting(self, setting: str) -> str:
        """Look up a Django setting (strings only)."""
        return self.settings[setting]

    def export_suite_signing_keys(self, suite_id: int) -> str:
        """Export signing keys for a suite as an ASCII-armored keyring."""
        return self.suite_signing_keys[suite_id]


class FakeTaskDatabase(FakeTaskDatabaseBase):
    """
    Fake task database with empty configure implementation.

    This is used for tests that do not perform database access
    """

    def configure(self, task: "BaseTask[Any, Any]") -> None:
        """Perform server-side task configuration."""
        raise NotImplementedError()


class TaskHelperMixin[TaskClass: BaseTask[Any, Any]]:
    """Helper mixin for Task tests."""

    task_class: type[TaskClass]
    task: TaskClass
    task_db: MockTaskDatabase

    def __init_subclass__(cls, **kwargs: Any) -> None:
        """Compute cls.task_class."""
        super().__init_subclass__(**kwargs)

        # The task class used for this test suite, computed by introspecting
        # the type argument used to specialize this generic class.
        [cls.task_class] = extract_generic_type_arguments(cls, TaskHelperMixin)

    def _create_task(self, task_data: dict[str, Any]) -> TaskClass:
        """
        Create a task with the given task data.

        This should not be used directly by tests, but may be overridden by
        variants of this mixin.
        """
        return self.task_class(task_data)

    def configure_task(
        self,
        task_data: dict[str, Any] | None = None,
        override: dict[str, Any] | None = None,
        remove: list[str] | None = None,
    ) -> None:
        """
        Run self.task.configure(task_data) with different inputs.

        Copy self.SAMPLE_TASK_DATA or task_data and modify it, call
        self.task.configure(modified_SAMPLE_TASK_DATA).

        :param task_data: if provided use this as a base dictionary. If None,
          use self.SAMPLE_TASK_DATA.
        :param override: dictionary with the keys to modify (only root keys,
          not recursive search) by its values. Modify or add them.
        :param remove: list of keys to remove.
        """
        if task_data is None:  # pragma: no cover
            task_data = cast(
                TaskTestProtocol[TaskClass], self
            ).SAMPLE_TASK_DATA.copy()
        if override:
            for key, value in override.items():
                task_data[key] = value
        if remove:
            for key in remove:
                task_data.pop(key, None)

        self.task = self._create_task(task_data)

    def compute_dynamic_data[DTD: BaseDynamicTaskData](
        self, task: BaseTask[Any, DTD]
    ) -> DTD:
        """Compute dynamic data for this task."""
        assert self.task_db is not None
        self.task_db.resolve_inputs(task)
        return task.build_dynamic_data()

    def set_dynamic_data(self) -> None:
        """
        Compute and store the task's dynamic data.

        ``task_db`` must be set up in advance with enough information to
        allow computing dynamic data.
        """
        self.task.dynamic_data = self.compute_dynamic_data(self.task)

    def write_os_release(
        self, filename: Path, codename: str = "bookworm", version: int = 12
    ) -> dict[str, str]:
        """
        Write in directory / self.task._OS_RELEASE_FILE a release file.

        :return: dictionary with keys to values.
        """
        data = {
            "ID": "debian",
            "NAME": "Debian GNU/Linux",
            "PRETTY_NAME": f"Debian GNU/Linux {version} ({codename})",
            "VERSION": f"{version} ({codename})",
            "VERSION_CODENAME": codename,
            "VERSION_ID": str(version),
        }

        if version >= 13:
            del data["VERSION"]
            del data["VERSION_ID"]

        filename.write_text(
            "\n".join(f"{k}={shlex.quote(v)}" for k, v in data.items()) + "\n"
        )
        return data


class ExternalTaskHelperMixin[ExternalTaskClass: BaseExternalTask[Any, Any]](
    TaskHelperMixin[ExternalTaskClass]
):
    """Helper mixin for testing Tasks that run on external workers."""

    task_class: type[ExternalTaskClass]
    task: ExternalTaskClass

    def mock_debusine(self: TaskTestProtocol[ExternalTaskClass]) -> MagicMock:
        """Create a Debusine mock and configure self.task for it. Return it."""
        debusine_mock = mock.create_autospec(spec=Debusine)
        assert isinstance(debusine_mock, MagicMock)

        # Worker would set the server
        self.task.configure_server_access(debusine_mock)

        return debusine_mock

    def mock_image_download(
        self,
        debusine_mock: MagicMock,
        create_image: bool = False,
        system_image: bool = False,
    ) -> ArtifactResponse:
        """
        Configure a fake ImageCache path and fake_system_*_artifact.

        The tarball itself isn't faked, unless create_image is set.

        If system_image is set, a system-image is faked instead of a tarball.
        """
        if system_image:
            artifact_response = self.fake_system_image_artifact()
        else:
            artifact_response = self.fake_system_tarball_artifact()
        debusine_mock.artifact_get.return_value = artifact_response

        def download_tarball(
            artifact_id: int,  # noqa: ARG001, U100
            filename: str,  # noqa: ARG001, U100
            destination: Path,
        ) -> None:
            with tarfile.open(destination, "w:xz") as f:
                tarinfo = tarfile.TarInfo("./")
                tarinfo.type = tarfile.DIRTYPE
                f.addfile(tarinfo)
                tarinfo = tarfile.TarInfo("./sbin")
                tarinfo.type = tarfile.DIRTYPE
                f.addfile(tarinfo)
                tarinfo = tarfile.TarInfo("./sbin/init")
                tarinfo.size = 3
                f.addfile(tarinfo, BytesIO(b"123"))

        def download_image(
            artifact_id: int,  # noqa: ARG001, U100
            filename: str,  # noqa: ARG001, U100
            destination: Path,
        ) -> None:
            with destination.open("wb") as f:
                f.write(b"qcow!")

        self.image_cache_path = Path(
            mkdtemp(prefix="debusine-testsuite-images-")
        )
        cast(TestCase, self).addCleanup(rmtree, self.image_cache_path)

        img_cache_patcher = mock.patch(
            "debusine.tasks.executors.images.ImageCache.image_cache_path",
            self.image_cache_path,
        )
        img_cache_patcher.start()
        cast(TestCase, self).addCleanup(img_cache_patcher.stop)

        patcher = mock.patch(
            "debusine.tasks.executors.images.ImageCache"
            "._download_image_artifact"
        )
        self.download_image_artifact_mock = patcher.start()
        cast(TestCase, self).addCleanup(patcher.stop)

        if create_image:
            if system_image:
                self.download_image_artifact_mock.side_effect = download_image
            else:
                self.download_image_artifact_mock.side_effect = download_tarball

        return artifact_response

    def patch_prepare_executor_instance(self) -> MagicMock:
        """
        Patch self.task._prepare_executor_instance(), return its mock.

        Side effect of self.task._prepare_executor_instance(): set
        self.task.executor and self.task.executor_instance.
        """
        patcher = mock.patch.object(
            self.task, "_prepare_executor_instance", autospec=True
        )
        mocked = patcher.start()

        def mock_executor_executor_instance() -> None:
            self.task.executor = MagicMock(spec=ExecutorInterface)
            self.task.executor_instance = MagicMock(spec=UnshareInstance)
            self.task.executor_instance.run.return_value.returncode = 0

        mocked.side_effect = mock_executor_executor_instance
        cast(TestCase, self).addCleanup(patcher.stop)
        return mocked

    def fake_system_tarball_artifact(self) -> ArtifactResponse:
        """Create a fake ArtifactResponse for a debian:system-tarball."""
        return create_artifact_response(
            id=42,
            workspace="Testing",
            category=ArtifactCategory.SYSTEM_TARBALL,
            created_at=dt.datetime(2024, 1, 1, 0, 0, 0, tzinfo=dt.UTC),
            data={
                "architecture": "amd64",
                "codename": "bookworm",
                "filename": "system.tar.xz",
                "vendor": "debian",
                "pkglist": {},
            },
            download_tar_gz_url="https://example.com/download-42/",
            files=FilesResponseType(
                {
                    "system.tar.xz": FileResponse(
                        size=4001,
                        checksums={"sha256": "abc123"},
                        type="file",
                        url="https://example.com/download-system.tar.xz",
                    ),
                }
            ),
            files_to_upload=[],
        )

    def fake_system_image_artifact(self) -> ArtifactResponse:
        """Create a fake ArtifactResponse for a debian:system-image."""
        return create_artifact_response(
            id=42,
            workspace="Testing",
            category=ArtifactCategory.SYSTEM_IMAGE,
            created_at=dt.datetime(2024, 1, 1, 0, 0, 0, tzinfo=dt.UTC),
            data={
                "architecture": "amd64",
                "codename": "bookworm",
                "filename": "image.qcow2",
                "vendor": "debian",
                "image_format": "qcow2",
                "filesystem": "ext4",
                "size": 2 * 1024 * 1024 * 1024,
                "boot_mechanism": "efi",
                "pkglist": {},
            },
            download_tar_gz_url="https://example.com/download-42/",
            files=FilesResponseType(
                {
                    "image.qcow2": FileResponse(
                        size=4001,
                        checksums={"sha256": "abc123"},
                        type="file",
                        url="https://example.com/download-image.qcow2",
                    ),
                }
            ),
            files_to_upload=[],
        )

    def fake_debian_source_package_artifact(self) -> ArtifactResponse:
        """Create a fake ArtifactResponse for a debian:source-package."""
        return create_artifact_response(
            id=6,
            workspace="Testing",
            category=ArtifactCategory.SOURCE_PACKAGE,
            created_at=dt.datetime(2024, 1, 1, 0, 0, 0, tzinfo=dt.UTC),
            data={
                "name": "foo",
                "version": "1.0-1",
                "type": "dpkg",
                "dsc_fields": {},
            },
            download_tar_gz_url="https://example.com/download-6/",
            files=FilesResponseType(
                {
                    "foo_1.0-1.dsc": FileResponse(
                        size=10,
                        checksums={"sha256": "abc123"},
                        type="file",
                        url="https://example.com/foo_1.0-1.dsc",
                    ),
                }
            ),
            files_to_upload=[],
        )

    def fake_debian_binary_package_artifact(self) -> ArtifactResponse:
        """Create a fake ArtifactResponse for a debian:binary-package."""
        return create_artifact_response(
            id=7,
            workspace="Testing",
            category=ArtifactCategory.BINARY_PACKAGE,
            created_at=dt.datetime(2024, 1, 1, 0, 0, 0, tzinfo=dt.UTC),
            data={
                "srcpkg_name": "foo",
                "srcpkg_version": "1.0-1",
                "type": "dpkg",
                "dsc_fields": {},
                "dsc_control_files": [],
            },
            download_tar_gz_url="https://example.com/download-7/",
            files=FilesResponseType(
                {
                    "foo_1.0-1_all.deb": FileResponse(
                        size=10,
                        checksums={"sha256": "abc123"},
                        type="file",
                        url="https://example.com/foo_1.0-1_all.deb",
                    ),
                }
            ),
            files_to_upload=[],
        )

    def fake_debian_upload_artifact(
        self,
        source_package_name: str = "foo",
        architecture: str = "source",
        names: list[str] | None = None,
    ) -> ArtifactResponse:
        """Create a fake ArtifactResponse for a debian:upload."""
        if names is None:
            names = [
                "{source_package_name}_1.0.orig.tar.xz",
                "{source_package_name}_1.0-1.debian.tar.xz",
                "{source_package_name}_1.0-1.dsc",
            ]
        return create_artifact_response(
            id=8,
            workspace="Testing",
            category=ArtifactCategory.UPLOAD,
            created_at=dt.datetime(2024, 1, 1, 0, 0, 0, tzinfo=dt.UTC),
            data={
                "type": "dpkg",
                "changes_fields": {
                    "Architecture": architecture,
                    "Files": [{"name": name} for name in names],
                },
            },
            download_tar_gz_url="https://example.com/download-8/",
            files=FilesResponseType(
                {
                    name: FileResponse(
                        size=len(name),
                        checksums={
                            "sha256": hashlib.sha256(name.encode()).hexdigest()
                        },
                        type="file",
                        url=f"https://example.com/{name}",
                    )
                    for name in [
                        f"{source_package_name}_1.0-1_{architecture}.changes",
                        *names,
                    ]
                }
            ),
            files_to_upload=[],
        )


class SampleBaseTask[TD: BaseTaskData, DTD: BaseDynamicTaskData](
    BaseTask[TD, DTD], metaclass=abc.ABCMeta
):
    """Common test implementation of abstract task methods."""

    @override
    def build_dynamic_data(self) -> DTD:
        """Resolve artifact lookups for this task."""
        return self.dynamic_task_data_type()


class SampleBaseExternalTask[TD: BaseTaskData, DTD: BaseDynamicTaskData](
    SampleBaseTask[TD, DTD], BaseExternalTask[TD, DTD], metaclass=abc.ABCMeta
):
    """Common test implementation of BaseExternalTask methods."""


class SampleBaseTaskWithExecutor[
    TDE: BaseTaskDataWithExecutor,
    DTDE: BaseDynamicTaskDataWithExecutor,
](SampleBaseTask[TDE, DTDE], BaseTaskWithExecutor[TDE, DTDE]):
    """Common test implementation of BaseExternalTask methods."""



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Signing task to sign the contents of an artifact."""

import shutil
from collections import defaultdict
from itertools import zip_longest
from pathlib import Path
from typing import Any, assert_never, override

from django.conf import settings

import debusine.worker.tags as wtags
from debusine import utils
from debusine.artifacts import SigningOutputArtifact
from debusine.artifacts.models import SigningResult, WorkRequestResults
from debusine.assets import AssetCategory, KeyPurpose
from debusine.client.models import (
    AssetPermissionCheckResponse,
    FilesResponseType,
    RelationType,
)
from debusine.signing.db.models import Key, sign
from debusine.signing.models import SigningMode
from debusine.signing.tasks import BaseSigningTask
from debusine.signing.tasks.models import SignData, SignDynamicData
from debusine.tasks.inputs import MultiInput


class Sign(BaseSigningTask[SignData, SignDynamicData]):
    """Task that signs data using a key."""

    TASK_VERSION = 1

    unsigned = MultiInput()

    @override
    def post_init(self) -> None:
        super().post_init()
        self._files: dict[int, FilesResponseType] | None = None
        self._key: Key | None = None
        self._permission_checks: dict[int, AssetPermissionCheckResponse] = {}
        self._results: dict[int, list[SigningResult]] = defaultdict(list)

    @override
    @classmethod
    def analyze_worker(cls) -> dict[str, Any]:
        """Report metadata for this task on this worker."""
        metadata = super().analyze_worker()

        metadata[cls.prefix_with_task_name("available:uefi")] = (
            utils.is_command_available("sbsign")
        )

        return metadata

    @override
    @classmethod
    def get_provided_worker_tags(cls) -> set[str]:
        tags = super().get_provided_worker_tags()
        if utils.is_command_available("sbsign"):
            tags.add(wtags.CAP_PREFIX + "sbsign")
        return tags

    @override
    def compute_system_required_tags(self) -> set[str]:
        tags = super().compute_system_required_tags()
        tags.add(wtags.CAP_PREFIX + "sbsign")
        return tags

    @override
    def can_run_on(self, worker_metadata: dict[str, Any]) -> bool:
        """Check if the specified worker can run the task."""
        if not super().can_run_on(worker_metadata):
            return False

        # OpenPGP support is implemented in Python and is always available.
        # Other key purposes require external tools.
        if self.data.purpose != KeyPurpose.OPENPGP_UPLOAD:
            available_key = self.prefix_with_task_name(
                f"available:{self.data.purpose}"
            )
            if not worker_metadata.get(available_key, False):
                return False

        return True

    @override
    def build_dynamic_data(self) -> SignDynamicData:
        """Resolve artifact lookups for this task."""
        unsigned = self.unsigned
        return SignDynamicData(
            unsigned_ids=unsigned.get_ids(),
            unsigned_binary_package_names=[
                getattr(item.data, "binary_package_name", None)
                for item in unsigned
            ],
        )

    @override
    def fetch_input(self, destination: Path) -> bool:
        """Fetch artifacts and database objects needed by the task."""
        assert self.debusine is not None
        assert self.dynamic_data is not None
        assert self.work_request_id is not None
        assert self.workspace_name is not None

        asset_slug = f"{self.data.purpose}:{self.data.key}"
        self._files = {}
        for unsigned_id in self.dynamic_data.unsigned_ids:
            unsigned = self.fetch_artifact(unsigned_id, destination)
            for trusted_cert in unsigned.data.get("trusted_certs") or []:
                if trusted_cert not in settings.DEBUSINE_SIGNING_TRUSTED_CERTS:
                    self.append_to_log_file(
                        "fetch_input.log",
                        [
                            f"This installation cannot sign objects that trust "
                            f"certificate '{trusted_cert}'"
                        ],
                    )
                    return False
            self._files[unsigned_id] = unsigned.files

            permission_check = self.debusine.asset_permission_check(
                asset_category=AssetCategory.SIGNING_KEY,
                asset_slug=asset_slug,
                permission_name="sign_with",
                artifact_id=unsigned_id,
                work_request_id=self.work_request_id,
                workspace=self.workspace_name,
            )
            if not permission_check.has_permission:
                self.append_to_log_file(
                    "fetch_input.log",
                    [
                        f"User {permission_check.username} does not have "
                        f"permission to sign {unsigned_id} with {asset_slug}."
                    ],
                )
                return False
            self._permission_checks[unsigned_id] = permission_check

        try:
            self._key = Key.objects.get(
                purpose=self.data.purpose, fingerprint=self.data.key
            )
        except Key.DoesNotExist:
            self.append_to_log_file(
                "fetch_input.log",
                [
                    f"Signing key {self.data.purpose}:{self.data.key} "
                    f"does not exist"
                ],
            )
            return False

        return True

    @override
    def configure_for_execution(
        self,
        download_directory: Path,  # noqa: U100
    ) -> bool:
        """Configure task variables."""
        # Nothing to do here; fetch_input does it all.
        return True

    @override
    def prepare_to_run(
        self, download_directory: Path, execute_directory: Path
    ) -> None:
        """Copy downloaded files into the execution directory."""
        shutil.copytree(download_directory, execute_directory / "input")

    @override
    def run(self, execute_directory: Path) -> WorkRequestResults:
        """Sign all the requested files."""
        assert self.work_request_id is not None
        assert self._files is not None
        assert self._key is not None

        with self.open_debug_log_file("cmd-output.log", mode="wb") as log_file:
            signed_any = False
            failed = False
            for unsigned_id, files in self._files.items():
                for file_path, file_response in files.items():
                    data_path = execute_directory / "input" / file_path
                    signature_path = (
                        execute_directory / "output" / f"{file_path}.sig"
                    )
                    # TODO: Implement kernel module signing with kmodsign
                    # (#939392)
                    if data_path.suffix == ".ko" and KeyPurpose.UEFI:
                        self._results[unsigned_id].append(
                            SigningResult(
                                file=file_path,
                                error_message=(
                                    "Skipped kernel module (not yet supported)"
                                ),
                            )
                        )
                        continue
                    signature_path.parent.mkdir(parents=True, exist_ok=True)
                    # TODO: The signing mode could be specified in the task
                    # data.  For now we hardcode ones that make sense.
                    match self.data.purpose:
                        case KeyPurpose.UEFI:
                            mode = SigningMode.DETACHED
                        case KeyPurpose.OPENPGP_UPLOAD:
                            mode = SigningMode.CLEAR
                        case KeyPurpose.OPENPGP_REPOSITORY:  # pragma: no cover
                            raise AssertionError(
                                f"Sign task does not support "
                                f"{KeyPurpose.OPENPGP_REPOSITORY} keys"
                            )
                        case _ as unreachable:
                            assert_never(unreachable)
                    permission_check = self._permission_checks[unsigned_id]
                    try:
                        sign(
                            [self._key],
                            data_path,
                            signature_path,
                            mode,
                            self.work_request_id,
                            username=permission_check.username,
                            user_id=permission_check.user_id,
                            resource=permission_check.resource,
                            log_file=log_file,
                        )
                        signed_any = True
                    except Exception as e:
                        failed = True
                        self._results[unsigned_id].append(
                            SigningResult(file=file_path, error_message=str(e))
                        )
                    else:
                        self._results[unsigned_id].append(
                            SigningResult(
                                file=file_path, output_file=f"{file_path}.sig"
                            )
                        )

        if signed_any and not failed:
            return WorkRequestResults.SUCCESS
        return WorkRequestResults.FAILURE

    @override
    def upload_artifacts(
        self,
        execute_directory: Path,
        *,
        execution_result: WorkRequestResults,  # noqa: U100
    ) -> None:
        """Upload artifacts for the task."""
        assert self.dynamic_data is not None
        assert self.work_request_id is not None
        assert self.debusine is not None
        assert self._key is not None

        for unsigned_id, unsigned_binary_package_name in zip_longest(
            self.dynamic_data.unsigned_ids,
            self.dynamic_data.unsigned_binary_package_names,
        ):
            signing_output_artifact = SigningOutputArtifact.create(
                KeyPurpose(self._key.purpose),
                self._key.fingerprint,
                self._results[unsigned_id],
                [
                    execute_directory / "output" / result.output_file
                    for result in self._results[unsigned_id]
                    if result.output_file is not None
                ],
                execute_directory / "output",
                binary_package_name=unsigned_binary_package_name,
            )
            uploaded_signing_output_artifact = self.debusine.upload_artifact(
                signing_output_artifact,
                workspace=self.workspace_name,
                work_request=self.work_request_id,
            )
            self.debusine.relation_create(
                uploaded_signing_output_artifact.id,
                unsigned_id,
                RelationType.RELATES_TO,
            )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Data models for db scopes."""

import enum
import re
from typing import Any, Self, TYPE_CHECKING, TypeAlias, cast

import pgtrigger
from django.core.exceptions import ImproperlyConfigured, ValidationError
from django.db import models
from django.db.models import CheckConstraint, F, Q, QuerySet, UniqueConstraint

from debusine.db.context import context
from debusine.db.models import permissions
from debusine.db.models.files import File, FileStore
from debusine.db.models.permissions import (
    Allow,
    PermissionUser,
    permission_check,
    permission_filter,
)

if TYPE_CHECKING:
    from django.http import HttpRequest
    from django_stubs_ext.db.models import TypedModelMeta

    from debusine.db.models.auth import User
    from debusine.server.file_backend.interface import FileBackendInterface
    from debusine.web.views.ui.scopes import ScopeUI
else:
    TypedModelMeta = object

#: Scope names reserved for use in toplevel URL path components
RESERVED_SCOPE_NAMES = frozenset(
    (
        "accounts",
        "admin",
        "api",
        "api-auth",
        "artifact",
        "task-status",
        "user",
        "workers",
        "work-request",
        "workspace",
    )
)

#: Regexp matching the structure of scope names
scope_name_regex = re.compile(r"^[A-Za-z][A-Za-z0-9+._-]*$")


def is_valid_scope_name(value: str) -> bool:
    """Check if value is a valid scope name."""
    if value in RESERVED_SCOPE_NAMES:
        return False
    return bool(scope_name_regex.match(value))


def validate_scope_name(value: str) -> None:
    """Validate scope names."""
    if not is_valid_scope_name(value):
        raise ValidationError(
            "%(value)r is not a valid scope name", params={"value": value}
        )


class ScopeRoleBase(permissions.RoleBase):
    """Scope role implementation."""

    implied_by_scope_roles: frozenset["ScopeRoles"]

    def _setup(self) -> None:
        """Set up implications for a newly constructed role."""
        for i in self.implied_by:
            raise ImproperlyConfigured(
                f"Scope roles do not support implications by {type(i)}"
            )
        self.implied_by_scope_roles = frozenset((cast(ScopeRoles, self),))

    def q(self, user: "User") -> Q:
        """Return a Q expression to select scopes with this role."""
        return Q(
            roles__group__users=user,
            roles__role__in=self.implied_by_scope_roles,
        )

    def implies(self, role: "ScopeRoles") -> bool:
        """Check if this role implies the given one."""
        return self in role.implied_by_scope_roles


class ScopeRoles(permissions.Roles, ScopeRoleBase, enum.ReprEnum):
    """Available roles for a Scope."""

    OWNER = "owner"


ScopeRoles.setup()


class ScopeQuerySet[A](QuerySet["Scope", A]):
    """Custom QuerySet for Scope."""

    def with_role(self, user: PermissionUser, role: ScopeRoles) -> Self:
        """Keep only resources where the user has the given role."""
        if not user or not user.is_authenticated:
            return self.none()
        return self.filter(
            pk__in=self.model.objects.filter(role.q(user)).distinct()
        )

    @permission_filter(workers=Allow.ALWAYS, anonymous=Allow.ALWAYS)
    def can_display(self, user: PermissionUser) -> Self:  # noqa: U100
        """Keep only Scopes that can be displayed."""
        assert user is not None  # Enforced by decorator
        return self

    @permission_filter()
    def can_create_workspace(self, user: PermissionUser) -> Self:
        """Keep only Scopes where the user can create workspaces."""
        return self.with_role(user, Scope.Roles.OWNER)


class ScopeManager(models.Manager["Scope"]):
    """Manager for Scope model."""

    def get_roles_model(self) -> type["ScopeRole"]:
        """Get the model used for role assignment."""
        return ScopeRole

    def get_queryset(self) -> ScopeQuerySet[Any]:
        """Use the custom QuerySet."""
        return ScopeQuerySet(self.model, using=self._db)


class Scope(models.Model):
    """
    Scope model.

    This is used to create different distinct sets of groups and workspaces
    """

    Roles: TypeAlias = ScopeRoles

    name = models.CharField(
        max_length=255,
        unique=True,
        validators=[validate_scope_name],
        help_text="internal name for the scope",
    )
    label = models.CharField(
        max_length=255,
        unique=True,
        help_text="User-visible name for the scope",
    )
    icon = models.CharField(
        max_length=255,
        default="",
        blank=True,
        help_text=(
            "Optional user-visible icon,"
            " resolved via ``{% static %}`` in templates"
        ),
    )
    file_stores = models.ManyToManyField(
        FileStore, related_name="scopes", through="db.FileStoreInScope"
    )

    objects = ScopeManager.from_queryset(ScopeQuerySet)()

    class Meta(TypedModelMeta):
        base_manager_name = "objects"

    def __str__(self) -> str:
        """Return basic information of Scope."""
        return self.name

    def ui(self, request: "HttpRequest") -> "ScopeUI":
        """Return a UI helper for this instance."""
        from debusine.web.views.ui.scopes import ScopeUI

        return ScopeUI.get(request, self)

    @permission_check(
        "{user} cannot display scope {resource}",
        workers=Allow.ALWAYS,
        anonymous=Allow.ALWAYS,
    )
    def can_display(self, user: PermissionUser) -> bool:  # noqa: ARG002, U100
        """Check if the scope can be displayed."""
        return True

    @permission_check(
        "{user} cannot create workspaces in {resource}",
    )
    def can_create_workspace(self, user: PermissionUser) -> bool:
        """Check if the user can create workspaces in this scope."""
        return self.has_role(user, Scope.Roles.OWNER)

    def has_role(self, user: PermissionUser, role: ScopeRoles) -> bool:
        """Check if the user has the given role on this Scope."""
        if not user or not user.is_authenticated:
            return False
        if context.user == user and context.scope == self:
            return any(r.implies(role) for r in context.scope_roles)
        return Scope.objects.with_role(user, role).filter(pk=self.pk).exists()

    # See https://github.com/typeddjango/django-stubs/issues/1047 for the typing
    def get_group_roles(
        self, user: PermissionUser
    ) -> QuerySet["ScopeRole", str]:
        """Get the roles of the user on this scope."""
        if not user or not user.is_authenticated:
            return ScopeRole.objects.none().values_list("role", flat=True)
        else:
            return (
                ScopeRole.objects.filter(resource=self, group__users=user)
                .values_list("role", flat=True)
                .distinct()
            )

    def upload_file_stores(
        self,
        fileobj: File,
        *,
        enforce_soft_limits: bool = False,
        include_write_only: bool = False,
    ) -> QuerySet[FileStore]:
        """
        Find the file stores in this scope where `fileobj` can be uploaded.

        The returned query set is in descending order of upload priority,
        breaking ties by ascending file store IDs.

        :param enforce_soft_limits: Enforce `soft_max_size` policies as well
          as `max_size`.
        :param include_write_only: Uploading a file to a write-only store
          will mean that debusine won't download it from there again, so
          that only makes sense in specialized situations such as populating
          a backup store.  Set this to True to include write-only stores in
          the returned query set.
        """
        file_stores = (
            self.file_stores.exclude(filestoreinscope__read_only=True)
            .exclude(max_size__lt=F("total_size") + fileobj.size)
            .order_by(
                F("filestoreinscope__upload_priority").desc(nulls_last=True),
                "id",
            )
        )
        if enforce_soft_limits:
            # TODO: We should also enforce FileStoreInScope.soft_max_size.
            file_stores = file_stores.exclude(
                soft_max_size__lt=F("total_size") + fileobj.size
            )
        if not include_write_only:
            file_stores = file_stores.exclude(filestoreinscope__write_only=True)
        return file_stores

    def upload_file_backend(self, fileobj: File) -> "FileBackendInterface[Any]":
        """
        Find the best file backend for uploading `fileobj`.

        :raises IndexError: if there is no such file backend.
        """
        return self.upload_file_stores(fileobj)[0].get_backend_object()

    def download_file_stores(self, fileobj: File) -> QuerySet[FileStore]:
        """
        Find the file stores in this scope that have `fileobj`, if any.

        The returned query set is in descending order of download priority,
        breaking ties in descending order of upload priority and then by
        ascending file store IDs.
        """
        return (
            self.file_stores.exclude(filestoreinscope__write_only=True)
            .filter(files=fileobj)
            .order_by(
                F("filestoreinscope__download_priority").desc(nulls_last=True),
                F("filestoreinscope__upload_priority").desc(nulls_last=True),
                "id",
            )
        )

    def download_file_backend(
        self, fileobj: File
    ) -> "FileBackendInterface[Any]":
        """
        Find the best file backend for downloading `fileobj`.

        :raises IndexError: if there is no such file backend.
        """
        return self.download_file_stores(fileobj)[0].get_backend_object()


class ScopeRole(models.Model):
    """Role assignments for scopes."""

    Roles: TypeAlias = ScopeRoles

    resource = models.ForeignKey(
        Scope,
        on_delete=models.CASCADE,
        related_name="roles",
    )

    group = models.ForeignKey(
        "Group",
        on_delete=models.CASCADE,
        related_name="scope_roles",
    )

    role = models.CharField(max_length=16, choices=Roles.choices)

    class Meta(TypedModelMeta):
        constraints = [
            UniqueConstraint(
                fields=["resource", "group", "role"],
                name="%(app_label)s_%(class)s_unique_resource_group_role",
            ),
        ]

    def __str__(self) -> str:
        """Return a description of the role assignment."""
        return f"{self.group}─{self.role}⟶{self.resource}"


class FileStoreInScope(models.Model):
    """Database model used for extra data on Scope/FileStore relations."""

    scope = models.ForeignKey(Scope, on_delete=models.PROTECT)
    file_store = models.ForeignKey(FileStore, on_delete=models.PROTECT)

    _file_store_instance_wide = models.BooleanField(
        default=True,
        editable=False,
        db_column="file_store_instance_wide",
        help_text="Synced from FileStore.instance_wide; do not update directly",
    )

    #: The priority of this store for the purpose of storing new files.
    #: When adding a new file, debusine tries stores whose policies allow
    #: adding new files in descending order of upload priority, counting
    #: null as the lowest.
    upload_priority = models.IntegerField(blank=True, null=True)

    #: The priority of this store for the purpose of serving files to
    #: clients.  When downloading a file, debusine tries stores in
    #: descending order of download priority, counting null as the lowest;
    #: it breaks ties in descending order of upload priority, again counting
    #: null as the lowest.  If there is still a tie, it picks one of the
    #: possibilities arbitrarily.
    download_priority = models.IntegerField(blank=True, null=True)

    #: If True, the storage maintenance job ensures that this store has a
    #: copy of all files in the scope.
    populate = models.BooleanField(default=False)

    #: If True, the storage maintenance job moves all files in this scope to
    #: some other store in the same scope, following the same rules for
    #: finding a target store as for uploads of new files.  It does not move
    #: into a store if that would take its total size over `soft_max_size`
    #: (either for the scope or the file store), and it logs an error if it
    #: cannot find any eligible target store.
    drain = models.BooleanField(default=False)

    #: If this field is set, then constrain `drain` to use the store with
    #: the given name in this scope.
    #
    # TODO: ruff is correct that we shouldn't use null=True on a TextField,
    # but fixing that retroactively is non-trivial.
    drain_to = models.TextField(blank=True, null=True)  # noqa: DJ001

    #: If True, debusine will not add new files to this store.  Use this in
    #: combination with `drain` to prepare for removing the file store.
    read_only = models.BooleanField(default=False)

    #: If True, debusine will not read files from this store.  This is
    #: suitable for provider storage classes that are designed for long-term
    #: archival rather than routine retrieval, such as S3 Glacier Deep
    #: Archive.
    write_only = models.BooleanField(default=False)

    #: An integer specifying the number of bytes that the file store can
    #: hold for this scope (accounting files that are in multiple scopes to
    #: all of the scopes in question).  This limit may be exceeded
    #: temporarily during uploads; the storage maintenance job will move the
    #: least-recently-used files to another file store to get back below the
    #: limit.
    soft_max_size = models.IntegerField(blank=True, null=True)

    class Meta(TypedModelMeta):
        triggers = [
            pgtrigger.Trigger(
                name="db_filestoreinscope_sync_instance_wide",
                operation=pgtrigger.Insert | pgtrigger.Update,
                when=pgtrigger.Before,
                func=" ".join(
                    """
                    NEW.file_store_instance_wide =
                        (SELECT db_filestore.instance_wide
                         FROM db_filestore
                         WHERE db_filestore.id = NEW.file_store_id);
                    RETURN NEW;
                    """.split()
                ),
            )
        ]
        constraints = [
            UniqueConstraint(
                fields=["scope", "file_store"],
                name="%(app_label)s_%(class)s_unique_scope_file_store",
            ),
            UniqueConstraint(
                fields=["file_store"],
                name=(
                    "%(app_label)s_%(class)s_"
                    "unique_file_store_not_instance_wide"
                ),
                condition=Q(_file_store_instance_wide=False),
            ),
            # It does not make sense to request a store to be populated
            # while also requesting it to be either drained or read-only.
            CheckConstraint(
                name="%(app_label)s_%(class)s_consistent_populate",
                check=Q(populate=False) | Q(drain=False, read_only=False),
            ),
        ]

    def __str__(self) -> str:
        """Return basic information of FileStoreInScope."""
        return f"{self.scope}/{self.file_store.name}"



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""UI helpers for collections."""

from functools import cached_property

from django.urls import reverse

from debusine.db.models import Collection, CollectionItem
from debusine.web.views.places import IndexPlace, ResourcePlace
from debusine.web.views.ui.base import UI


class CollectionUI(UI[Collection]):
    """UI helpers for Collection instances."""

    @cached_property
    def place(self) -> ResourcePlace:
        """Return a place to show this collection."""
        place = ResourcePlace(
            title=f"{self.instance.name}@{self.instance.category} collection",
            breadcrumb=self.instance.name,
            url=self.instance.get_absolute_url(),
            description=f"{self.instance.name}@{self.instance.category}"
            " collection",
        )
        place.parent = self.instance.workspace.ui(
            self.request
        ).place_collection_list
        return place

    @cached_property
    def place_search(self) -> IndexPlace:
        """Return a place to search this collection."""
        return IndexPlace(
            title=f"Search {self.instance.name}@{self.instance.category}",
            url=self.instance.get_absolute_url_search(),
            description=f"{self.instance} collection search",
            parent=self.instance.workspace.ui(
                self.request
            ).place_collection_list,
        )

    @cached_property
    def place_update(self) -> IndexPlace:
        """Return a place to configure this collection."""
        return IndexPlace(
            title="Configure",
            url=reverse(
                "workspaces:collections:update",
                kwargs={
                    "wname": self.instance.workspace.name,
                    "cname": self.instance.name,
                    "ccat": self.instance.category,
                },
            ),
            parent=self.place,
        )

    @cached_property
    def place_delete(self) -> IndexPlace:
        """Return a place to delete this collection."""
        return IndexPlace(
            title="Delete",
            url=reverse(
                "workspaces:collections:delete",
                kwargs={
                    "wname": self.instance.workspace.name,
                    "cname": self.instance.name,
                    "ccat": self.instance.category,
                },
            ),
            parent=self.place,
        )


class CollectionItemUI(UI[CollectionItem]):
    """UI helpers for CollectionItem instances."""

    @cached_property
    def place(self) -> ResourcePlace:
        """Return a place to show this collection item."""
        return ResourcePlace(
            title=self.instance.name,
            breadcrumb=self.instance.name,
            url=self.instance.get_absolute_url(),
            parent=self.instance.parent_collection.ui(self.request).place,
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for DebusineTaskConfiguration collections."""

import unittest
from typing import Any, ClassVar
from unittest import mock

from debusine.artifacts.models import (
    BareDataCategory,
    DebusineTaskConfiguration,
    TaskTypes,
)
from debusine.db.models import Collection, CollectionItem
from debusine.db.playground import scenarios
from debusine.server.collections import DebusineTaskConfigurationManager
from debusine.server.collections.base import ItemAdditionError
from debusine.server.collections.debusine_task_configuration import (
    apply_configuration,
    build_configuration,
    list_configuration,
    lookup_config_by_name,
    lookup_templates,
)
from debusine.tasks import TaskConfigError
from debusine.test.django import TestCase


class TaskConfigTestCase(TestCase):
    """Base class for tests that apply task configuration."""

    scenario = scenarios.DefaultContext()
    collection: ClassVar[Collection]
    manager: ClassVar[DebusineTaskConfigurationManager]

    @classmethod
    def setUpTestData(cls) -> None:
        """Set up common test data."""
        super().setUpTestData()
        cls.collection = cls.scenario.default_task_configuration_collection
        cls.manager = DebusineTaskConfigurationManager(
            collection=cls.collection
        )

    @classmethod
    def _add_config(cls, entry: DebusineTaskConfiguration) -> CollectionItem:
        """Add a config entry to config_collection."""
        return cls.manager.add_bare_data(
            BareDataCategory.TASK_CONFIGURATION,
            user=cls.scenario.user,
            data=entry,
        )


class DebusineTaskConfigurationManagerTests(TaskConfigTestCase):
    """Tests for DebusineTaskConfigurationManager."""

    def test_add_bare_data_no_data(self) -> None:
        """`add_bare_data` requires item data."""
        with self.assertRaisesRegex(
            ItemAdditionError,
            "Adding to debusine:task-configuration requires data",
        ):
            self.manager.add_bare_data(
                BareDataCategory.TASK_CONFIGURATION,
                user=self.scenario.user,
            )

    def test_add_bare_data_refuses_name(self) -> None:
        """`add_bare_data` refuses an explicitly-specified name."""
        with self.assertRaisesRegex(
            ItemAdditionError,
            "Cannot use an explicit item name when adding to "
            "debusine:task-configuration",
        ):
            self.manager.add_bare_data(
                BareDataCategory.TASK_CONFIGURATION,
                user=self.scenario.user,
                data=DebusineTaskConfiguration(
                    task_type=TaskTypes.WORKER,
                    task_name="sbuild",
                    subject="base-files",
                    context="sid",
                ),
                name="override",
            )

    def test_add_bare_data_raise_item_addition_error(self) -> None:
        """`add_bare_data` raises an error for duplicate names."""
        data = DebusineTaskConfiguration(template="test")
        self.manager.add_bare_data(
            BareDataCategory.TASK_CONFIGURATION,
            user=self.scenario.user,
            data=data,
        )

        with self.assertRaisesRegex(
            ItemAdditionError, "db_collectionitem_unique_active_name"
        ):
            self.manager.add_bare_data(
                BareDataCategory.TASK_CONFIGURATION,
                user=self.scenario.user,
                data=data,
            )

    def test_add_bare_data_replace(self) -> None:
        """`add_bare_data` can replace an existing bare data item."""
        data = DebusineTaskConfiguration(template="test")

        item_old = self.manager.add_bare_data(
            BareDataCategory.TASK_CONFIGURATION,
            user=self.scenario.user,
            data=data,
        )

        item_new = self.manager.add_bare_data(
            BareDataCategory.TASK_CONFIGURATION,
            user=self.scenario.user,
            data=data,
            replace=True,
        )

        serialized_data = data.model_dump(mode="json", exclude_unset=True)

        item_old.refresh_from_db()
        self.assertEqual(item_old.name, "template:test")
        self.assertEqual(item_old.child_type, CollectionItem.Types.BARE)
        self.assertEqual(item_old.data, serialized_data)
        self.assertEqual(item_old.removed_by_user, self.scenario.user)
        self.assertIsNotNone(item_old.removed_at)
        self.assertEqual(item_new.name, "template:test")
        self.assertEqual(item_new.child_type, CollectionItem.Types.BARE)
        self.assertEqual(item_new.data, serialized_data)
        self.assertIsNone(item_new.removed_at)

    def test_add_bare_data_replace_nonexistent(self) -> None:
        """Replacing a nonexistent bare data item is allowed."""
        data = DebusineTaskConfiguration(template="test")

        item = self.manager.add_bare_data(
            BareDataCategory.TASK_CONFIGURATION,
            user=self.scenario.user,
            data=data,
            replace=True,
        )

        self.assertEqual(item.name, "template:test")
        self.assertEqual(item.child_type, CollectionItem.Types.BARE)
        self.assertEqual(
            item.data, data.model_dump(mode="json", exclude_unset=True)
        )

    def test_remove_item_bare_data(self) -> None:
        """`remove_item` removes a bare data item."""
        data = DebusineTaskConfiguration(template="test")
        item = self.manager.add_bare_data(
            BareDataCategory.TASK_CONFIGURATION,
            user=self.scenario.user,
            data=data,
        )

        self.manager.remove_item(item, user=self.scenario.user)

        self.assertEqual(item.removed_by_user, self.scenario.user)
        self.assertIsNotNone(item.removed_at)

    def test_lookup_unexpected_format(self) -> None:
        """`lookup` raises `LookupError` for an unexpected format."""
        with self.assertRaisesRegex(
            LookupError, r'^Unexpected lookup format: "foo:bar"'
        ):
            self.manager.lookup("foo:bar")

    def test_lookup_return_none(self) -> None:
        """`lookup` returns None if there are no matches."""
        self.assertIsNone(self.manager.lookup("name:does-not-exist"))

    def test_lookup_by_name(self) -> None:
        """Lookup by name works."""
        data = DebusineTaskConfiguration(template="test")
        item = self.manager.add_bare_data(
            BareDataCategory.TASK_CONFIGURATION,
            user=self.scenario.user,
            data=data,
        )
        self.assertEqual(self.manager.lookup("name:template:test"), item)


class ConfigLookupTests(TaskConfigTestCase):
    """Tests for the lookup_* methods."""

    def test_lookup_config_by_name_successful(self) -> None:
        entry = DebusineTaskConfiguration(
            task_type=TaskTypes.WORKER, task_name="noop"
        )
        self._add_config(entry)
        self.assertEqual(
            lookup_config_by_name(self.collection, entry.name()),
            entry,
        )

    def test_lookup_config_by_name_with_deleted_entries(self) -> None:
        entry = DebusineTaskConfiguration(
            task_type=TaskTypes.WORKER, task_name="noop"
        )
        self.manager.remove_item(
            self._add_config(entry), user=self.scenario.user
        )
        self._add_config(entry)
        self.assertEqual(
            lookup_config_by_name(self.collection, entry.name()),
            entry,
        )

    def test_lookup_config_by_name_missing(self) -> None:
        """Lookup returns None if the named item does not exist."""
        self.assertIsNone(
            lookup_config_by_name(self.collection, "Worker:noop::")
        )

    def test_lookup_config_by_name_corrupted(self) -> None:
        """Lookup raises if the item contains invalid data."""
        entry = DebusineTaskConfiguration(
            task_type=TaskTypes.WORKER, task_name="noop"
        )
        item = self._add_config(entry)
        item.data["invalid_field"] = "value"
        item.save()
        with self.assertRaisesRegex(
            TaskConfigError,
            r"(?s)debusine:task-configuration item does not validate"
            r".+Extra inputs are not permitted",
        ):
            lookup_config_by_name(self.collection, entry.name())

    def test_lookup_templates_empty(self) -> None:
        """Lookup templates with use_templates is empty."""
        entry = DebusineTaskConfiguration(
            task_type=TaskTypes.WORKER, task_name="noop"
        )
        self.assertEqual(
            list(lookup_templates(self.collection, entry)),
            [],
        )

    def test_lookup_templates_missing_item(self) -> None:
        """Lookup templates with use_templates is empty."""
        entry = DebusineTaskConfiguration(
            task_type=TaskTypes.WORKER,
            task_name="noop",
            use_templates=["missing"],
        )
        with self.assertRaisesRegex(
            TaskConfigError,
            r"Worker:noop:: references missing template 'missing'",
        ):
            list(lookup_templates(self.collection, entry))

    def test_lookup_templates_found(self) -> None:
        """Lookup templates with one template."""
        template = DebusineTaskConfiguration(
            template="template",
        )
        self._add_config(template)

        entry = DebusineTaskConfiguration(
            task_type=TaskTypes.WORKER,
            task_name="noop",
            use_templates=["template"],
        )
        self.assertEqual(
            list(lookup_templates(self.collection, entry)),
            [template],
        )

    def test_lookup_templates_diamond(self) -> None:
        """Lookup templates with a diamond dependency."""
        base = DebusineTaskConfiguration(
            template="base",
        )
        self._add_config(base)
        a = DebusineTaskConfiguration(template="a", use_templates=["base"])
        self._add_config(a)
        b = DebusineTaskConfiguration(template="b", use_templates=["base"])
        self._add_config(b)
        ab = DebusineTaskConfiguration(template="ab", use_templates=["a", "b"])
        self._add_config(ab)

        entry = DebusineTaskConfiguration(
            task_type=TaskTypes.WORKER,
            task_name="noop",
            use_templates=["ab"],
        )
        self.assertEqual(
            list(lookup_templates(self.collection, entry)),
            [base, a, base, b, ab],
        )

    def test_lookup_templates_cycle(self) -> None:
        """Lookup templates with a dependency cycle."""
        a = DebusineTaskConfiguration(template="a", use_templates=["b"])
        self._add_config(a)
        b = DebusineTaskConfiguration(template="b", use_templates=["a"])
        self._add_config(b)

        entry = DebusineTaskConfiguration(
            task_type=TaskTypes.WORKER,
            task_name="noop",
            use_templates=["a"],
        )
        with self.assertRaisesRegex(
            TaskConfigError,
            r"template:b: template lookup cycle detected: a→b→a",
        ):
            list(lookup_templates(self.collection, entry))

    def test_list_configuration_no_templates(self) -> None:
        """list_configuration with no templates."""
        entry = DebusineTaskConfiguration(
            task_type=TaskTypes.WORKER, task_name="noop"
        )
        self._add_config(entry)
        self.assertEqual(
            list(list_configuration(self.collection, entry.name())),
            [entry],
        )

    def test_list_configuration_missing_item(self) -> None:
        """list_configuration with a missing names yields no entries."""
        self.assertEqual(
            list(list_configuration(self.collection, "Worker:noop::")),
            [],
        )

    def test_list_configuration_templates_first(self) -> None:
        """list_configuration yields templates before items."""
        a = DebusineTaskConfiguration(
            template="a",
        )
        entry = DebusineTaskConfiguration(
            task_type=TaskTypes.WORKER,
            task_name="noop",
            use_templates=["a"],
        )
        self._add_config(entry)
        with mock.patch(
            "debusine.server.collections"
            ".debusine_task_configuration.lookup_templates",
            return_value=iter((a, a, a)),
        ):
            self.assertEqual(
                list(list_configuration(self.collection, entry.name())),
                [entry, a, a, a],
            )


class BuildConfigurationTests(unittest.TestCase):
    """Test :py:func:build_configuration`."""

    def entry(
        self, name: str = "test", **kwargs: Any
    ) -> DebusineTaskConfiguration:
        return DebusineTaskConfiguration(template=name, **kwargs)

    def test_empty(self) -> None:
        """Run with no entries."""
        self.assertEqual(build_configuration(()), ({}, {}))

    def test_one(self) -> None:
        """Run with one entry."""
        entry = self.entry(
            delete_values=["val0"],
            default_values={"val1": 1},
            override_values={"val2": 2},
            lock_values=["val3"],
        )
        self.assertEqual(
            build_configuration([entry]),
            ({"val1": 1}, {"val2": 2}),
        )

    def test_default_and_override(self) -> None:
        """Run with one entry with same default and override keys."""
        entry = self.entry(
            default_values={"val1": 1},
            override_values={"val1": 11},
        )
        self.assertEqual(
            build_configuration([entry]),
            ({"val1": 1}, {"val1": 11}),
        )

    def test_stack(self) -> None:
        """Run with two entries."""
        entry1 = self.entry(
            default_values={"val1": 1, "base1": 1},
            override_values={"val1": 2, "base2": 2},
        )
        entry2 = self.entry(
            default_values={"val1": 11},
            override_values={"val1": 22},
        )
        self.assertEqual(
            build_configuration([entry1, entry2]),
            ({"val1": 11, "base1": 1}, {"val1": 22, "base2": 2}),
        )

    def test_delete_values(self) -> None:
        """Test delete_values."""
        entry1 = self.entry(
            default_values={"val1": 1, "val2": 2},
            override_values={"val1": 11, "val2": 22},
        )
        entry2 = self.entry(
            delete_values=["val1"],
        )
        self.assertEqual(
            build_configuration([entry1, entry2]),
            ({"val2": 2}, {"val2": 22}),
        )

    def test_delete_and_set_values(self) -> None:
        """Test deleting and then setting values."""
        entry1 = self.entry(
            default_values={"val1": 1, "val2": 2},
            override_values={"val1": 11, "val2": 22},
        )
        entry2 = self.entry(
            delete_values=["val1"],
            default_values={"val1": 111},
            override_values={"val1": 112},
        )
        self.assertEqual(
            build_configuration([entry1, entry2]),
            ({"val1": 111, "val2": 2}, {"val1": 112, "val2": 22}),
        )

    def test_lock_values(self) -> None:
        """Test locked_values."""
        entry1 = self.entry(
            default_values={"val1": 1, "val2": 2},
            override_values={"val1": 11, "val2": 22},
            lock_values=["val1"],
        )
        entry2 = self.entry(
            delete_values=["val1", "val2"],
            default_values={"val1": 111, "val3": 3},
            override_values={"val1": 112, "val3": 33},
        )
        self.assertEqual(
            build_configuration([entry1, entry2]),
            ({"val1": 1, "val3": 3}, {"val1": 11, "val3": 33}),
        )


class ApplyConfigurationTests(TaskConfigTestCase):
    """Test :py:func:apply_configuration`."""

    def test_no_config(self) -> None:
        """Test without configuration to apply."""
        task_data: dict[str, Any] = {}
        apply_configuration(
            task_data, self.collection, TaskTypes.WORKER, "noop", None, None
        )
        self.assertEqual(task_data, {"task_configuration": self.collection.pk})

    def test_with_config(self) -> None:
        """Test with overrides to apply."""
        entry = DebusineTaskConfiguration(
            task_type=TaskTypes.WORKER,
            task_name="noop",
            default_values={"default": "added"},
            override_values={"override": "overridden"},
        )
        self._add_config(entry)

        task_data: dict[str, Any] = {
            "key": "value",
            "override": "original",
        }
        apply_configuration(
            task_data, self.collection, TaskTypes.WORKER, "noop", None, None
        )
        self.assertEqual(
            task_data,
            {
                "key": "value",
                "override": "overridden",
                "default": "added",
                "task_configuration": self.collection.pk,
            },
        )

    def test_defaults_with_none(self) -> None:
        """Test setting defaults on existing values set to None."""
        entry = DebusineTaskConfiguration(
            task_type=TaskTypes.WORKER,
            task_name="noop",
            default_values={"default": "added"},
        )
        self._add_config(entry)

        task_data: dict[str, Any] = {"default": None}
        apply_configuration(
            task_data, self.collection, TaskTypes.WORKER, "noop", None, None
        )
        self.assertEqual(
            task_data,
            {"default": "added", "task_configuration": self.collection.pk},
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""URLs related to work_requests."""

from django.urls import path

import debusine.web.views.work_request as views

app_name = "work_requests"

urlpatterns = [
    path(
        "",
        views.WorkRequestListView.as_view(),
        name="list",
    ),
    path(
        "<int:pk>/",
        views.WorkRequestDetailView.as_view(),
        name="detail",
    ),
    path(
        "<int:pk>/retry/",
        views.WorkRequestRetryView.as_view(),
        name="retry",
    ),
    path(
        "<int:pk>/abort/",
        views.WorkRequestAbortView.as_view(),
        name="abort",
    ),
    path(
        "<int:pk>/unblock/",
        views.WorkRequestUnblockView.as_view(),
        name="unblock",
    ),
    path(
        "<int:pk>/confirm/",
        views.WorkRequestConfirmView.as_view(),
        name="confirm",
    ),
    path(
        "create/",
        views.WorkRequestCreateView.as_view(),
        name="create",
    ),
]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for OpenMetrics statistics in Debusine."""

import datetime as dt
from collections.abc import Iterator
from contextlib import contextmanager
from unittest import mock

from django.test import SimpleTestCase, override_settings
from django.utils import timezone
from prometheus_client.core import Metric

from debusine.artifacts import LocalArtifact
from debusine.artifacts.models import ArtifactCategory, TaskTypes
from debusine.assets.models import AssetCategory
from debusine.db.models import (
    Artifact,
    Identity,
    User,
    WorkRequest,
    WorkerPoolStatistics,
)
from debusine.server.open_metrics import DebusineCollector, ZeroGauge
from debusine.server.signon.providers import Provider
from debusine.tasks.models import BackendType, NoopData, WorkerType
from debusine.test.django import TestCase


class ZeroGaugeTests(SimpleTestCase):
    """Tests for our zero-filling ZeroGauge."""

    def test_add_metric_stores_samples(self) -> None:
        gauge = ZeroGauge("test", "Description", labels=["type"])
        gauge.add_metric(["foo"], 1)

        self.assertEqual(len(gauge.samples), 1)
        sample = gauge.samples[0]
        self.assertEqual(sample.value, 1)
        self.assertEqual(sample.labels, {"type": "foo"})

    def test_add_metric_sets_seen(self) -> None:
        gauge = ZeroGauge("test", "Description", labels=["type"])
        gauge.add_metric(["foo"], 1)
        gauge.add_metric(["bar"], 2)

        self.assertEqual(gauge._seen, {("foo",), ("bar",)})

    def test_fill_zeros(self) -> None:
        gauge = ZeroGauge("test", "Description", labels=["type"])
        gauge.add_metric(["foo"], 1)
        gauge.fill_zeros(("foo", "bar"))

        self.assertEqual(len(gauge.samples), 2)
        sample = gauge.samples[1]
        self.assertEqual(sample.value, 0)
        self.assertEqual(sample.labels, {"type": "bar"})

    def test_fill_zeros_complete(self) -> None:
        gauge = ZeroGauge("test", "Description", labels=["type"])
        gauge.add_metric(["foo"], 1)
        gauge.add_metric(["bar"], 2)
        gauge.fill_zeros(("foo", "bar"))

        self.assertEqual(len(gauge.samples), 2)
        self.assertNotEqual(gauge.samples[0].value, 0)
        self.assertNotEqual(gauge.samples[1].value, 0)

    def test_fill_zeros_empty(self) -> None:
        gauge = ZeroGauge("test", "Description", labels=["type"])
        gauge.fill_zeros(("foo", "bar"))

        self.assertEqual(len(gauge.samples), 2)
        self.assertEqual(gauge.samples[0].value, 0)
        self.assertEqual(gauge.samples[1].value, 0)

    def test_fill_zeros_matrix(self) -> None:
        gauge = ZeroGauge("test", "Description", labels=("a", "b"))
        gauge.add_metric(("0", "0"), 1)
        gauge.fill_zeros(("0", "1"), ("0", "1"))

        self.assertEqual(len(gauge.samples), 4)


class DebusineCollectorTests(TestCase):
    """Test DebusineCollector's collectors."""

    def collector(self) -> DebusineCollector:
        collector = DebusineCollector()
        return collector

    def sample_by_label(self, metric: Metric, labels: dict[str, str]) -> float:
        """Find the sample for labels in metric."""
        for sample in metric.samples:
            if sample.labels == labels:
                return sample.value
        raise KeyError(f"{labels} not found in {metric}")

    def assertMetricEquals(
        self, metric: Metric, value: float, **labels: str
    ) -> None:
        """Assert that there is a sample for labels in metric with value."""
        self.assertEqual(self.sample_by_label(metric, labels), value)

    def assertNoMetric(self, metric: Metric, **labels: str) -> None:
        """Assert that there is no sample for labels in metric."""
        with self.assertRaises(KeyError):
            self.sample_by_label(metric, labels)

    def assertNoNullLabels(self, metric: Metric) -> None:
        """Assert that all metric labels have string values."""
        for sample in metric.samples:
            for key in sample.labels:
                self.assertIsNotNone(
                    sample.labels[key], msg=f"Null value for label '{key}'"
                )

    def test_count_by_category(self) -> None:
        self.playground.create_artifact()

        gauge = self.collector().count_by_category(
            Artifact, LocalArtifact.artifact_categories()
        )
        self.assertEqual(gauge.name, "artifacts")
        self.assertEqual(gauge.documentation, "Number of Artifacts")

        scope = self.playground.get_default_scope().name
        self.assertMetricEquals(
            gauge, 1, category=ArtifactCategory.TEST, scope=scope
        )
        self.assertMetricEquals(
            gauge, 0, category=ArtifactCategory.BLHC, scope=scope
        )

    def test_count_assets(self) -> None:
        self.playground.create_signing_key_asset()

        gauge = self.collector().count_assets()
        self.assertEqual(gauge.name, "assets")

        scope = self.playground.get_default_scope().name
        self.assertMetricEquals(
            gauge, 1, category=AssetCategory.SIGNING_KEY, scope=scope
        )
        self.assertNoMetric(
            gauge, category=AssetCategory.CLOUD_PROVIDER_ACCOUNT, scope=scope
        )
        self.assertMetricEquals(
            gauge, 0, category=AssetCategory.CLOUD_PROVIDER_ACCOUNT, scope=""
        )
        self.assertNoMetric(gauge, category=AssetCategory.SIGNING_KEY, scope="")

    def test_measure_file_stores(self) -> None:
        store = self.playground.get_default_file_store()
        contents = b"abc123"
        self.playground.create_file_in_backend(contents=contents)
        gauges = {
            gauge.name: gauge
            for gauge in self.collector().measure_file_stores()
        }
        self.assertEqual(len(gauges), 2)

        file_store_size = gauges["file_store_size"]
        self.assertMetricEquals(
            file_store_size,
            len(contents),
            backend=store.backend,
            name=store.name,
        )
        file_store_max_size = gauges["file_store_max_size"]
        self.assertNoMetric(
            file_store_max_size, backend=store.backend, name=store.name
        )

    def test_measure_file_stores_max_size(self) -> None:
        store = self.playground.get_default_file_store()
        store.max_size = 1024
        store.save()
        gauges = {
            gauge.name: gauge
            for gauge in self.collector().measure_file_stores()
        }
        file_store_max_size = gauges["file_store_max_size"]
        self.assertMetricEquals(
            file_store_max_size, 1024, backend=store.backend, name=store.name
        )

    def test_count_groups(self) -> None:
        scope1 = self.playground.get_default_scope()
        scope2 = self.playground.get_or_create_scope("scope2")
        self.playground.create_group("group1", scope=scope1)
        self.playground.create_group("group2", scope=scope2, ephemeral=True)
        gauge = self.collector().count_groups()
        self.assertEqual(gauge.name, "groups")
        self.assertMetricEquals(gauge, 1, ephemeral="0", scope=scope1.name)
        self.assertMetricEquals(gauge, 0, ephemeral="0", scope=scope2.name)
        self.assertMetricEquals(gauge, 0, ephemeral="1", scope=scope1.name)
        self.assertMetricEquals(gauge, 1, ephemeral="1", scope=scope2.name)

    def test_count_tokens(self) -> None:
        self.playground.create_bare_token(enabled=False)
        self.playground.create_worker()  # creates a worker token
        self.playground.create_user_token()
        self.playground.create_user_token()

        gauge = self.collector().count_tokens()
        self.assertEqual(gauge.name, "tokens")
        self.assertMetricEquals(gauge, 1, user="0", worker="0", enabled="0")
        self.assertMetricEquals(gauge, 1, user="0", worker="1", enabled="1")
        self.assertMetricEquals(gauge, 2, user="1", worker="0", enabled="1")
        self.assertMetricEquals(gauge, 0, user="1", worker="0", enabled="0")

    def test_count_users(self) -> None:
        self.playground.get_default_user()
        user2 = self.playground.create_user("user2")
        user2.is_active = False
        user2.save()
        gauge = self.collector().count_users()
        self.assertEqual(gauge.name, "users")
        self.assertMetricEquals(gauge, 1, active="1")
        self.assertMetricEquals(gauge, 1, active="0")

    @contextmanager
    def freeze_time(self) -> Iterator[None]:
        frozen_time = timezone.now()
        with mock.patch("django.utils.timezone.now", return_value=frozen_time):
            yield

    def test_count_user_activity(self) -> None:
        with self.freeze_time():
            for username, ages in (
                ("user1", (0, 365)),
                ("user2", (4, 5)),
                ("user3", (364,)),
                ("user4", (365 * 2,)),
                ("user5", ()),
            ):
                user = self.playground.create_user(username)
                for days in ages:
                    wr = self.playground.create_workflow(created_by=user)
                    wr.created_at = timezone.now() - dt.timedelta(days=days)
                    wr.save()

            gauge = self.collector().count_user_activity()
        self.assertEqual(gauge.name, "user_activity")

        scope = self.playground.get_default_scope().name
        self.assertMetricEquals(gauge, 1, scope=scope, le="1")
        self.assertMetricEquals(gauge, 1, scope=scope, le="3")
        self.assertMetricEquals(gauge, 2, scope=scope, le="7")
        self.assertMetricEquals(gauge, 2, scope=scope, le="14")
        self.assertMetricEquals(gauge, 2, scope=scope, le="30")
        self.assertMetricEquals(gauge, 2, scope=scope, le="90")
        self.assertMetricEquals(gauge, 3, scope=scope, le="365")
        self.assertMetricEquals(gauge, 4, scope=scope, le="+Inf")

    @override_settings(SIGNON_PROVIDERS=[Provider(name="salsa", label="Salsa")])
    def test_count_user_identities_activity(self) -> None:
        with self.freeze_time():
            for username, issuer, ages in (
                ("user1", "salsa", (0, 365)),
                ("user2", None, (4, 5)),
                ("user3", "salsa", (364,)),
                ("user5", "salsa", ()),
                (None, "salsa", (45,)),
            ):
                user: User | None = None
                subject = "ex-user"
                if username:
                    user = self.playground.create_user(username)
                    subject = username
                if issuer:
                    Identity.objects.create(
                        user=user, issuer=issuer, subject=subject, claims={}
                    )
                for days in ages:
                    wr = self.playground.create_workflow(created_by=user)
                    wr.created_at = timezone.now() - dt.timedelta(days=days)
                    wr.save()

            gauge = self.collector().count_user_identities_activity()
        self.assertEqual(gauge.name, "user_identities_activity")

        scope = self.playground.get_default_scope().name
        self.assertMetricEquals(gauge, 1, scope=scope, issuer="salsa", le="1")
        self.assertMetricEquals(gauge, 1, scope=scope, issuer="salsa", le="3")
        self.assertMetricEquals(gauge, 1, scope=scope, issuer="salsa", le="7")
        self.assertMetricEquals(gauge, 1, scope=scope, issuer="salsa", le="14")
        self.assertMetricEquals(gauge, 1, scope=scope, issuer="salsa", le="30")
        self.assertMetricEquals(gauge, 1, scope=scope, issuer="salsa", le="90")
        self.assertMetricEquals(gauge, 2, scope=scope, issuer="salsa", le="365")
        self.assertMetricEquals(
            gauge, 2, scope=scope, issuer="salsa", le="+Inf"
        )
        self.assertNoMetric(gauge, scope=scope, issuer="", le="1")
        self.assertNoNullLabels(gauge)

    @override_settings(SIGNON_PROVIDERS=[Provider(name="salsa", label="Salsa")])
    def test_count_user_identities(self) -> None:
        user = self.playground.get_default_user()
        Identity.objects.create(
            user=user, issuer="salsa", subject="foo", claims={}
        )
        Identity.objects.create(
            user=None, issuer="salsa", subject="bar", claims={}
        )
        gauge = self.collector().count_user_identities()
        self.assertEqual(gauge.name, "user_identities")
        self.assertMetricEquals(gauge, 1, active="1", issuer="salsa")
        self.assertMetricEquals(gauge, 0, active="0", issuer="salsa")

    def test_count_work_requests(self) -> None:
        self.playground.create_worker_task(task_name="noop", mark_pending=True)
        self.playground.create_signing_task(task_name="noop", mark_running=True)
        self.playground.create_worker_task(
            task_name="noop",
            task_data=NoopData(
                build_architecture="amd64",
                backend=BackendType.UNSHARE,
            ),
            mark_running=True,
        )
        gauge = self.collector().count_work_requests()
        self.assertEqual(gauge.name, "work_requests")
        scope = self.playground.get_default_scope().name
        self.assertMetricEquals(
            gauge,
            1,
            task_type=TaskTypes.WORKER,
            task_name="noop",
            scope=scope,
            status=WorkRequest.Statuses.PENDING,
            build_architecture="",
            host_architecture="",
            backend="",
        )
        self.assertMetricEquals(
            gauge,
            1,
            task_type=TaskTypes.SIGNING,
            task_name="noop",
            scope=scope,
            status=WorkRequest.Statuses.RUNNING,
            build_architecture="",
            host_architecture="",
            backend="",
        )
        self.assertMetricEquals(
            gauge,
            1,
            task_type=TaskTypes.WORKER,
            task_name="noop",
            scope=scope,
            status=WorkRequest.Statuses.RUNNING,
            build_architecture="amd64",
            host_architecture="amd64",
            backend=BackendType.UNSHARE,
        )
        self.assertMetricEquals(
            gauge,
            0,
            task_type=TaskTypes.WORKER,
            task_name="noop",
            scope=scope,
            status=WorkRequest.Statuses.RUNNING,
            build_architecture="amd64",
            host_architecture="amd64",
            backend=BackendType.QEMU,
        )
        self.assertNoNullLabels(gauge)

    def test_count_workers(self) -> None:
        self.playground.create_worker(
            worker_type=WorkerType.EXTERNAL,
            extra_dynamic_metadata={
                "system:native_architecture": "amd64",
                "system:architectures": ["amd64", "i386"],
            },
        )
        worker_pool = self.playground.create_worker_pool("test")
        worker2 = self.playground.create_worker(
            worker_type=WorkerType.EXTERNAL,
            worker_pool=worker_pool,
            extra_dynamic_metadata={
                "system:native_architecture": "i386",
                "system:architectures": ["amd64", "x32"],
            },
        )
        # Overrides dynamic:
        worker2.static_metadata = {"system:native_architecture": "amd64"}
        worker2.save()

        worker3 = self.playground.create_worker(
            worker_type=WorkerType.SIGNING,
        )
        worker3.mark_connected()
        self.playground.create_worker_task(
            task_name="noop", mark_running=True, worker=worker3
        )

        gauge = self.collector().count_workers()
        self.assertEqual(gauge.name, "workers")
        self.assertMetricEquals(
            gauge,
            1,
            connected="0",
            busy="0",
            worker_type=WorkerType.EXTERNAL,
            worker_pool="",
            native_architecture="amd64",
            architecture_amd64="1",
            architecture_i386="1",
            architecture_x32="0",
        )
        self.assertMetricEquals(
            gauge,
            1,
            connected="0",
            busy="0",
            worker_type=WorkerType.EXTERNAL,
            worker_pool=worker_pool.name,
            native_architecture="amd64",
            architecture_amd64="1",
            architecture_i386="0",
            architecture_x32="1",
        )
        self.assertMetricEquals(
            gauge,
            0,
            connected="0",
            busy="0",
            worker_type=WorkerType.EXTERNAL,
            worker_pool="",
            native_architecture="i386",
            architecture_amd64="0",
            architecture_i386="0",
            architecture_x32="0",
        )
        self.assertMetricEquals(
            gauge,
            1,
            connected="1",
            busy="1",
            worker_type=WorkerType.SIGNING,
            worker_pool="",
            native_architecture="",
            architecture_amd64="0",
            architecture_i386="0",
            architecture_x32="0",
        )
        self.assertMetricEquals(
            gauge,
            0,
            connected="0",
            busy="0",
            worker_type=WorkerType.SIGNING,
            worker_pool="",
            native_architecture="",
            architecture_amd64="0",
            architecture_i386="0",
            architecture_x32="0",
        )
        self.assertNoNullLabels(gauge)

    def test_count_worker_pool_runtime(self) -> None:
        with self.freeze_time():
            worker_pool = self.playground.create_worker_pool("test")
            worker_pool_2 = self.playground.create_worker_pool(
                "test2", provider_account=worker_pool.provider_account
            )

            worker = self.playground.create_worker(worker_pool=worker_pool)
            worker.instance_created_at = timezone.now() - dt.timedelta(
                seconds=120
            )
            worker.save()
            worker2 = self.playground.create_worker(worker_pool=worker_pool)
            worker2.instance_created_at = timezone.now() - dt.timedelta(
                seconds=60
            )
            worker2.save()
            # Not part of the pool:
            worker3 = self.playground.create_worker()
            worker3.instance_created_at = timezone.now() - dt.timedelta(
                seconds=9999
            )
            worker3.save()

            WorkerPoolStatistics.objects.create(
                worker_pool=worker_pool,
                worker=worker,
                runtime=30,
            )
            WorkerPoolStatistics.objects.create(
                worker_pool=worker_pool,
                runtime=15,
            )
            WorkerPoolStatistics.objects.create(
                worker_pool=worker_pool_2,
                runtime=42,
            )

            gauge = self.collector().count_worker_pool_runtime()
        self.assertEqual(gauge.name, "worker_pool_runtime")
        self.assertMetricEquals(gauge, 120 + 15 + 60 + 30, worker_pool="test")
        self.assertMetricEquals(gauge, 42, worker_pool="test2")

    def test_count_workflow_templates(self) -> None:
        self.playground.create_workflow_template(name="test1", task_name="noop")
        gauge = self.collector().count_workflow_templates()
        self.assertEqual(gauge.name, "workflow_templates")
        scope = self.playground.get_default_scope().name
        self.assertMetricEquals(gauge, 1, task_name="noop", scope=scope)
        self.assertMetricEquals(gauge, 0, task_name="sbuild", scope=scope)

    def test_count_workspaces(self) -> None:
        self.playground.get_default_workspace()
        self.playground.create_workspace(name="private", public=False)
        scope2 = self.playground.get_or_create_scope("scope2")
        self.playground.create_workspace(
            name="expiring",
            expiration_delay=dt.timedelta(days=14),
            scope=scope2,
        )

        gauge = self.collector().count_workspaces()
        self.assertEqual(gauge.name, "workspaces")
        scope = self.playground.get_default_scope().name
        self.assertMetricEquals(gauge, 1, private="0", expires="0", scope=scope)
        self.assertMetricEquals(gauge, 1, private="1", expires="0", scope=scope)
        self.assertMetricEquals(
            gauge, 1, private="0", expires="1", scope=scope2.name
        )
        self.assertMetricEquals(
            gauge, 0, private="0", expires="0", scope=scope2.name
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""debusine collection views."""

from functools import cached_property
from typing import Any, override

from django.db.models import Count, Q, QuerySet
from django.http import Http404, HttpResponse, HttpResponseRedirect
from django.urls import reverse

from debusine.db.context import context
from debusine.db.models import Collection, CollectionItem
from debusine.web.forms import (
    CollectionCreateForm,
    CollectionSearchForm,
    CollectionUpdateForm,
    ModelFormBase,
)
from debusine.web.views import sidebar, ui_shortcuts
from debusine.web.views.base import (
    CreateViewBase,
    DeleteViewBase,
    DetailViewBase,
    FormMixinBase,
    ListViewBase,
    UpdateViewBase,
    WorkspaceView,
)
from debusine.web.views.mixins import RightbarUIMixin, UIShortcutsMixin
from debusine.web.views.places import Place, ResourcePlace
from debusine.web.views.table import TableMixin
from debusine.web.views.tables import CollectionItemTable
from debusine.web.views.view_utils import highlight_as_yaml


class CollectionListView(WorkspaceView, ListViewBase[Collection]):
    """List collections."""

    model = Collection
    template_name = "web/collection-list.html"
    context_object_name = "collection_list"
    ordering = ["category", "name"]

    def get_place(self) -> Place:
        """Return the Place for this page."""
        return (
            context.require_workspace().ui(self.request).place_collection_list
        )

    def get_queryset(self) -> QuerySet[Collection]:
        """Filter collection by accessible workspace."""
        queryset = super().get_queryset()
        return queryset.filter(workspace=context.workspace).exclude(
            category="debusine:workflow-internal"
        )


class CollectionCategoryListView(WorkspaceView, ListViewBase[Collection]):
    """List collections with a given category."""

    model = Collection
    template_name = "web/collection-category-list.html"
    context_object_name = "collection_list"
    ordering = ["name"]

    def get_place(self) -> Place:
        """Return the Place for this page."""
        workspace = context.require_workspace()
        return ResourcePlace(
            title=f"{workspace.name} {self.kwargs['ccat']} collections",
            url=reverse(
                "workspaces:collections:category_list",
                kwargs={"wname": workspace.name, "ccat": self.kwargs["ccat"]},
            ),
            breadcrumb=self.kwargs["ccat"],
        )

    def get_queryset(self) -> QuerySet[Collection]:
        """Filter collection by accessible workspace."""
        queryset = super().get_queryset()
        return (
            queryset.filter(workspace=context.workspace)
            .exclude(category="debusine:workflow-internal")
            .filter(category=self.kwargs["ccat"])
        )

    def get_context_data(self, *args: Any, **kwargs: Any) -> dict[str, Any]:
        """Return context_data with category."""
        context = super().get_context_data(**kwargs)
        context["category"] = self.kwargs["ccat"]
        return context


class CollectionCreateView(
    WorkspaceView, CreateViewBase[Collection, CollectionCreateForm]
):
    """Create a collection."""

    form_class = CollectionCreateForm
    template_name = "web/collection-create.html"

    @override
    def init_view(self) -> None:
        """Enforce permissions."""
        super().init_view()
        self.enforce(context.require_workspace().can_create_collections)

    @override
    def get_place(self) -> Place:
        return (
            context.require_workspace().ui(self.request).place_collection_create
        )


class CollectionViewMixin(WorkspaceView, UIShortcutsMixin, RightbarUIMixin):
    """Common functions for collection-specific views."""

    @cached_property
    def collection(self) -> Collection:
        """Collection for this request."""
        try:
            return Collection.objects.can_display(context.user).get(
                workspace=context.workspace,
                name=self.kwargs["cname"],
                category=self.kwargs["ccat"],
            )
        except Collection.DoesNotExist:
            raise Http404(
                f"{self.kwargs['cname']}@{self.kwargs['ccat']}"
                " collection not found"
            )

    def get_sidebar_items(self) -> list[sidebar.SidebarItem]:
        """Return a list of sidebar items."""
        items = super().get_sidebar_items()
        items.append(sidebar.create_collection(self.collection))
        assert context.workspace is not None
        items.append(sidebar.create_workspace(context.workspace))
        return items

    def get_context_data(self, *args: Any, **kwargs: Any) -> dict[str, Any]:
        """Return context_data with collection."""
        context = super().get_context_data(**kwargs)
        context["collection"] = self.collection
        return context


class CollectionDetailView(CollectionViewMixin, DetailViewBase[Collection]):
    """Show a collection detail."""

    model = Collection
    template_name = "web/collection-detail.html"
    context_object_name = "collection"

    def get_place(self) -> Place:
        """Return the Place for this page."""
        return self.object.ui(self.request).place

    def get_object(
        self,
        queryset: QuerySet[Collection] | None = None,  # noqa: ARG002, U100
    ) -> Collection:
        """Return the collection object to show."""
        return self.collection

    def get_context_data(self, *args: Any, **kwargs: Any) -> dict[str, Any]:
        """Return context_data with work_request_list and workspace_list."""
        context = super().get_context_data(**kwargs)
        if self.object.data:
            context["data"] = highlight_as_yaml(self.object.data)

        context["collections"] = (
            self.object.child_items.active()
            .filter(collection__isnull=False)
            .order_by("category", "name")
        )

        context["artifacts"] = (
            self.object.child_items.filter(artifact__isnull=False)
            .values("category")
            .annotate(
                count=Count("category", filter=Q(removed_at__isnull=True)),
                count_removed=Count(
                    "category", filter=Q(removed_at__isnull=False)
                ),
            )
            .order_by("category")
        )

        context["bare"] = (
            self.object.child_items.filter(
                collection__isnull=True, artifact__isnull=True
            )
            .values("category")
            .annotate(
                count=Count("category", filter=Q(removed_at__isnull=True)),
                count_removed=Count(
                    "category", filter=Q(removed_at__isnull=False)
                ),
            )
            .order_by("category")
        )

        return context


class CollectionUpdateView(
    CollectionViewMixin, UpdateViewBase[Collection, CollectionUpdateForm]
):
    """Edit configuration for a collection."""

    form_class = CollectionUpdateForm
    template_name = "web/collection-update.html"

    @override
    def get_place(self) -> Place:
        return self.object.ui(self.request).place_update

    def get_object(
        self,
        queryset: QuerySet[Collection] | None = None,  # noqa: ARG002, U100
    ) -> Collection:
        """Return the collection object to update."""
        self.enforce(self.collection.can_configure)
        return self.collection


class CollectionDeleteView(
    CollectionViewMixin, DeleteViewBase[Collection, ModelFormBase[Collection]]
):
    """Delete a collection."""

    object: Collection
    model = Collection
    template_name = "web/collection-confirm_delete.html"

    @override
    def get_place(self) -> Place:
        return self.object.ui(self.request).place_delete

    def get_object(
        self,
        queryset: QuerySet[Collection] | None = None,  # noqa: ARG002, U100
    ) -> Collection:
        """Return the collection object to delete."""
        self.enforce(self.collection.can_delete)
        return self.collection

    @override
    def get_context_data(self, *args: Any, **kwargs: Any) -> dict[str, Any]:
        ctx = super().get_context_data(**kwargs)
        item_stats = list(self.object.get_item_stats())
        ctx["items_total"] = sum(x["count"] for x in item_stats)
        ctx["item_stats"] = item_stats
        return ctx

    @override
    def form_valid(
        self,
        form: ModelFormBase[Collection],  # noqa: U100
    ) -> HttpResponse:
        success_url = self.get_success_url()
        self.object.delete_collection()
        return HttpResponseRedirect(success_url)

    @override
    def get_success_url(self) -> str:
        return context.require_workspace().get_absolute_url()


class CollectionSearchView(
    TableMixin[CollectionItem],
    CollectionViewMixin,
    FormMixinBase[CollectionSearchForm],
    ListViewBase[CollectionItem],
):
    """Search a collection contents."""

    model = CollectionItem
    template_name = "web/collection-search.html"
    context_object_name = "item_list"
    form_class = CollectionSearchForm
    table_class = CollectionItemTable
    paginate_by = 50

    def get_place(self) -> Place:
        """Return the Place for this page."""
        return self.collection.ui(self.request).place_search

    def get_queryset(self) -> QuerySet[CollectionItem]:
        """Filter items based on the search form."""
        queryset = (
            super().get_queryset().filter(parent_collection=self.collection)
        )
        queryset = queryset.select_related(
            "parent_collection__workspace__scope",
            "collection",
            "artifact__workspace__scope",
        )
        # FIXME: this builds the form twice: if it becomes a problem we can
        # override get_form to cache its result
        form = self.get_form()
        if form.is_valid():
            if name := form.cleaned_data["name"]:
                queryset = queryset.filter(name__startswith=name)
            if category := form.cleaned_data["category"]:
                queryset = queryset.filter(category=category)
            if form.cleaned_data["historical"]:
                queryset = queryset.filter(removed_at__isnull=False)
            else:
                queryset = queryset.filter(removed_at__isnull=True)
        else:
            queryset = queryset.none()
        return queryset

    def get_form_kwargs(self) -> dict[str, Any]:
        """Get arguments used to instantiate the form."""
        kwargs = super().get_form_kwargs()
        kwargs["instance"] = self.collection
        kwargs["data"] = self.request.GET
        return kwargs

    def get_context_data(self, *args: Any, **kwargs: Any) -> dict[str, Any]:
        """Return context_data with work_request_list and workspace_list."""
        context = super().get_context_data(**kwargs)
        context["collection"] = self.collection
        # TODO: we need to match model objects exactly in order to be able to
        # lookup UI shortcuts, and this is a hack to do that, and also avoid to
        # iterate queryset twice.
        # See !1540 for a possible way to avoid this hack
        page_obj = context["paginator"].page_obj
        page_obj.object_list = list(page_obj.object_list)
        for item in page_obj.object_list:
            self.add_object_ui_shortcuts(
                item, ui_shortcuts.create_collection_item(item)
            )
            if item.artifact is not None:
                self.add_object_ui_shortcuts(
                    item,
                    ui_shortcuts.create_artifact_view(item.artifact),
                    ui_shortcuts.create_artifact_download(item.artifact),
                )
        return context


class CollectionItemDetailView(
    CollectionViewMixin, DetailViewBase[CollectionItem]
):
    """Show a collection item detail."""

    model = CollectionItem
    template_name = "web/collection-item-detail.html"
    context_object_name = "item"

    def get_place(self) -> Place:
        """Return the Place for this page."""
        return self.object.ui(self.request).place

    def get_sidebar_items(self) -> list[sidebar.SidebarItem]:
        """Return a list of sidebar items."""
        items = super().get_sidebar_items()
        # TODO: create collection-specific or artifact-specific UI shortcuts as
        # needed
        items.append(
            sidebar.create_user(
                self.object.created_by_user, context=self.object
            )
        )
        items.append(sidebar.create_created_at(self.object.created_at))
        return items

    def get_object(
        self, queryset: QuerySet[CollectionItem] | None = None
    ) -> CollectionItem:
        """Return the collection object to show."""
        assert queryset is None
        queryset = self.get_queryset()
        try:
            return queryset.filter(removed_at__isnull=True).get(
                pk=self.kwargs["iid"], parent_collection=self.collection
            )
        except CollectionItem.DoesNotExist:
            raise Http404(
                f"{self.kwargs['iid']} ({self.kwargs['iname']})"
                " item not found in "
                f"{self.kwargs['cname']}@{self.kwargs['ccat']}"
            )

    def get_context_data(self, *args: Any, **kwargs: Any) -> dict[str, Any]:
        """Return context_data."""
        context = super().get_context_data(**kwargs)
        if self.object.data:
            context["data"] = highlight_as_yaml(self.object.data)
        return context



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for table filters."""

import re
from typing import Any, cast
from unittest import mock

import lxml
from django.core.exceptions import FieldError, ImproperlyConfigured
from django.db.models import Q
from django.forms import ChoiceField, Form, modelform_factory

from debusine.db.models import User
from debusine.test.django import TestCase
from debusine.web.views.table import (
    Column,
    FilterField,
    FilterSelectOne,
    FilterText,
    FilterToggle,
    Table,
)
from debusine.web.views.table.filters import (
    BoundFilter,
    BoundFilterField,
    BoundFilterSelectOne,
    MainEntryField,
    MainEntrySelectOne,
    MainEntryText,
    MainEntryToggle,
    MenuEntry,
    MenuEntryToggle,
    SelectOption,
)
from debusine.web.views.table.tests.test_table import TableTestCase
from debusine.web.views.tests.utils import ViewTestMixin


class MenuEntryTests(ViewTestMixin, TestCase):
    """Tests for :py:class:`MenuEntry`."""

    def mock_filter(self) -> BoundFilter[Any]:
        """Build a mock BoundFilter."""
        res = mock.Mock()
        res.filter_prefix = "fieldname"
        res.label = "Label"
        res.dom_id = "dom_id"
        res.dom_id_form_dropdown = "dom_id1"
        return res

    def test_render(self) -> None:
        """Test :py:func:`render`."""
        widget = MenuEntry(self.mock_filter())
        tree = self.assertHTMLValid(self.render_widget(widget))
        a = self.assertHasElement(tree, "body/a")
        self.assertEqual(a.get("data-bs-target"), "#dom_id1")
        self.assertTextContentEqual(a, "Label")


class MenuEntryToggleTests(ViewTestMixin, TestCase):
    """Tests for :py:class:`MenuEntryToggle`."""

    def mock_filter(self, value: Any) -> BoundFilter[Any]:
        """Build a mock BoundFilter."""
        res = mock.Mock()
        res.filter_prefix = "fieldname"
        res.label = "Label"
        res.qs_remove = ""
        res.qs_with_value = mock.Mock(return_value="fieldname=1")
        res.value = value
        return res

    def test_render_inactive(self) -> None:
        """Test :py:func:`render` on an inactive filter."""
        widget = MenuEntryToggle(self.mock_filter(None))
        tree = self.assertHTMLValid(self.render_widget(widget))
        a = self.assertHasElement(tree, "body/a")
        self.assertEqual(a.get("href"), "?fieldname=1")
        self.assertEqual(a.get("class"), "dropdown-item")
        self.assertIsNone(a.get("aria-current"))
        self.assertTextContentEqual(a, "Label")

    def test_render_active(self) -> None:
        """Test :py:func:`render` on an active filter."""
        widget = MenuEntryToggle(self.mock_filter(True))
        tree = self.assertHTMLValid(self.render_widget(widget))
        a = self.assertHasElement(tree, "body/a")
        self.assertEqual(a.get("href"), "?")
        self.assertEqual(a.get("class"), "dropdown-item active")
        self.assertEqual(a.get("aria-current"), "true")
        self.assertTextContentEqual(a, "Label")


class MainEntryFormTests(TableTestCase, ViewTestMixin):
    """Base class for main entry tests for entries with forms."""

    def mock_filter(self, value: Any = None) -> BoundFilterField[Any]:
        """Build a mock BoundFilter."""
        qstring: dict[str, Any] = {"order": "test"}
        if value:
            qstring["filter-username"] = value

        table = self._table(qstring=qstring)
        return cast(BoundFilterField[Any], table.filters["username"])

    def assertDropdownOpener(
        self, opener: lxml.objectify.ObjectifiedElement, f: BoundFilter[Any]
    ) -> None:
        """Check the attributes of the dropdown opener."""
        self.assertEqual(opener.get("id"), f"{f.dom_id}-form-dropdown")
        self.assertEqual(opener.get("data-bs-toggle"), "dropdown")
        self.assertEqual(opener.get("data-bs-auto-close"), "outside")

    def assertDropdown(
        self, dropdown: lxml.objectify.ObjectifiedElement, f: BoundFilter[Any]
    ) -> lxml.objectify.ObjectifiedElement:
        """
        Check the general layout of the dropdown form.

        :returns: the form element
        """
        btn_add = dropdown.button[0]
        self.assertEqual(btn_add.get("type"), "submit")
        self.assertEqual(btn_add.get("form"), f"{f.dom_id}-form")
        self.assertTextContentEqual(btn_add, "Add filter")

        btn_cancel = dropdown.button[1]
        self.assertElementHasClass(btn_cancel, "debusine-dropdown-close")
        self.assertTextContentEqual(btn_cancel, "Cancel")

        form = dropdown.form[0]
        self.assertEqual(form.get("id"), f"{f.dom_id}-form")
        self.assertEqual(form.get("method"), "get")
        return form

    def assertActive(
        self, tree: lxml.objectify.ObjectifiedElement, f: BoundFilter[Any]
    ) -> lxml.objectify.ObjectifiedElement:
        """
        Check that the widget is representing an active field.

        :returns: the form element
        """
        main = self.assertHasElement(tree, "body/div")
        buttongroup = self.assertHasElement(main, "div")
        self.assertEqual(
            buttongroup.get("aria-label"),
            f"Controls for active filter {f.label}",
        )

        opener = self.assertHasElement(buttongroup, "button")
        self.assertDropdownOpener(opener, f)
        self.assertTextContentEqual(opener, f"{f.label}: {f.value_formatted}")

        remover = self.assertHasElement(buttongroup, "a")
        self.assertEqual(remover.get("href"), "?order=test")
        self.assertEqual(remover.get("title"), "Remove filter")

        dropdown = self.assertHasElement(buttongroup, "div")
        return self.assertDropdown(dropdown, f)


class MainEntryFieldTests(MainEntryFormTests):
    """Tests for :py:class:`MainEntryField`."""

    def get_table_class(self) -> type[Table[User]]:
        class _Table(Table[User]):
            user = Column("User", ordering="username")
            filter_username = FilterField("Username")
            default_order = "user"

        return _Table

    def assertForm(
        self, f: BoundFilterField[Any], form: lxml.objectify.ObjectifiedElement
    ) -> None:
        """Check the form contents."""
        label = self.assertHasElement(form, "label")
        self.assertTextContentEqual(label, f.label)
        inp = self.assertHasElement(form, "input")
        self.assertEqual(inp.get("id"), f.form_field.id_for_label)
        self.assertEqual(inp.get("name"), f.filter_prefix)
        self.assertEqual(inp.get("value"), f.value)

    def test_render_active(self) -> None:
        """Test :py:func:`render` with an active field."""
        f = self.mock_filter("testvalue")
        widget = MainEntryField(f)
        tree = self.assertHTMLValid(self.render_widget(widget))
        form = self.assertActive(tree, f)
        self.assertForm(f, form)

    def test_render_inactive(self) -> None:
        """Test :py:func:`render` with an active field."""
        f = self.mock_filter()
        widget = MainEntryField(f)
        tree = self.assertHTMLValid(self.render_widget(widget))
        main = self.assertHasElement(tree, "body/div")
        self.assertDropdownOpener(main.div[0], f)
        form = self.assertDropdown(main.div[1], f)
        self.assertForm(f, form)


class MainEntryTextTests(MainEntryFormTests):
    """Tests for :py:class:`MainEntryText`."""

    def get_table_class(self) -> type[Table[User]]:
        class _Table(Table[User]):
            user = Column("User", ordering="username")
            filter_username = FilterText("Username")
            default_order = "user"

        return _Table

    def assertForm(
        self, f: BoundFilter[Any], form: lxml.objectify.ObjectifiedElement
    ) -> None:
        """Check the form contents."""
        label = self.assertHasElement(form, "label")
        self.assertTextContentEqual(label, f.label)
        inp = self.assertHasElement(form, "input")
        self.assertEqual(
            inp.get("id"), f"{f.dom_id}-formfield-{f.filter_prefix}"
        )
        self.assertEqual(inp.get("name"), f.filter_prefix)
        self.assertEqual(
            inp.get("value"), f.value if f.value is not None else ""
        )

    def test_render_active(self) -> None:
        """Test :py:func:`render` with an active field."""
        f = self.mock_filter("testvalue")
        widget = MainEntryText(f)
        tree = self.assertHTMLValid(self.render_widget(widget))
        form = self.assertActive(tree, f)
        self.assertForm(f, form)

    def test_render_inactive(self) -> None:
        """Test :py:func:`render` with an active field."""
        f = self.mock_filter()
        widget = MainEntryText(f)
        tree = self.assertHTMLValid(self.render_widget(widget))
        main = self.assertHasElement(tree, "body/div")
        self.assertDropdownOpener(main.div[0], f)
        form = self.assertDropdown(main.div[1], f)
        self.assertForm(f, form)


class MainEntrySelectOneTests(MainEntryFormTests):
    """Tests for :py:class:`MainEntrySelectOne`."""

    def get_table_class(self) -> type[Table[User]]:
        class _Table(Table[User]):
            user = Column("User", ordering="username")
            filter_username = FilterSelectOne(
                "Username",
                options=(
                    SelectOption("foo", "Foo", Q(username="foo")),
                    SelectOption("bar", "Bar", Q(username="bar")),
                ),
            )
            default_order = "user"

        return _Table

    def assertForm(
        self,
        f: BoundFilter[Any],
        form: lxml.objectify.ObjectifiedElement,
        active: int | None = None,
    ) -> None:
        """Check the form contents."""
        label = self.assertHasElement(form, "label")
        self.assertTextContentEqual(label, f.label)
        select = self.assertHasElement(form, "select")
        self.assertEqual(
            select.get("id"), f"{f.dom_id}-formfield-{f.filter_prefix}"
        )
        self.assertEqual(select.get("name"), f.filter_prefix)

        self.assertEqual(select.option[0].get("value"), "foo")
        self.assertEqual(bool(select.option[0].get("selected")), active == 0)
        self.assertTextContentEqual(select.option[0], "Foo")
        self.assertEqual(select.option[1].get("value"), "bar")
        self.assertEqual(bool(select.option[1].get("selected")), active == 1)
        self.assertTextContentEqual(select.option[1], "Bar")

    def test_render_active(self) -> None:
        """Test :py:func:`render` with an active field."""
        f = self.mock_filter("foo")
        widget = MainEntrySelectOne(f)
        tree = self.assertHTMLValid(self.render_widget(widget))
        form = self.assertActive(tree, f)
        self.assertForm(f, form, active=0)

    def test_render_inactive(self) -> None:
        """Test :py:func:`render` with an active field."""
        f = self.mock_filter()
        widget = MainEntrySelectOne(f)
        tree = self.assertHTMLValid(self.render_widget(widget))
        main = self.assertHasElement(tree, "body/div")
        self.assertDropdownOpener(main.div[0], f)
        form = self.assertDropdown(main.div[1], f)
        self.assertForm(f, form)


class MainEntryToggleTests(MainEntryFormTests):
    """Tests for :py:class:`MainEntryToggle`."""

    def get_table_class(self) -> type[Table[User]]:
        class _Table(Table[User]):
            user = Column("User", ordering="username")
            filter_username = FilterToggle("Username", q=Q(username="foo"))
            default_order = "user"

        return _Table

    def test_render_active(self) -> None:
        """Test :py:func:`render` with an active field."""
        f = self.mock_filter(True)
        widget = MainEntryToggle(f)
        tree = self.assertHTMLValid(self.render_widget(widget))
        div = self.assertHasElement(tree, "body/div")
        self.assertEqual(
            div.get("aria-label"), "Controls for active filter Username"
        )
        self.assertTextContentEqual(div.span, "Username")

        el_close = self.assertHasElement(div, "a")
        self.assertEqual(el_close.get("href"), "?order=test")

    def test_render_inactive(self) -> None:
        """Test :py:func:`render` with an active field."""
        f = self.mock_filter()
        widget = MainEntryToggle(f)
        self.assertEqual(self.render_widget(widget).strip(), "")


class FilterTextTests(TestCase):
    """Tests for :py:class:`FilterText`."""

    @classmethod
    def setUpTestData(cls) -> None:
        """Initialize class data."""
        super().setUpTestData()
        cls.playground.create_user("foo")
        cls.playground.create_user("foobar")
        cls.playground.create_user("bar")

    def test_metadata_default(self) -> None:
        """Test defaults for field metadata."""
        f = FilterText("User")
        f.name = "username"
        self.assertEqual(f.label, "User")
        self.assertEqual(f.q_field, "username__contains")
        self.assertIsNone(f.icon)
        self.assertIsNone(f.placeholder)
        self.assertIs(f.menu_entry, MenuEntry)
        self.assertIs(f.main_entry, MainEntryText)

    def test_metadata_full(self) -> None:
        """Test field metadata."""
        f = FilterText(
            "User",
            q_field="username__in",
            icon="icon",
            placeholder="placeholder",
        )
        self.assertEqual(f.label, "User")
        self.assertEqual(f.icon, "icon")
        self.assertEqual(f.q_field, "username__in")
        self.assertEqual(f.placeholder, "placeholder")

    def test_filter_queryset(self) -> None:
        """Test filter_queryset."""
        f = FilterText("User")
        f.name = "username"
        for value, expected in (
            ("foo", ["foo", "foobar"]),
            ("bar", ["foobar", "bar"]),
            ("barfoo", []),
        ):
            with self.subTest(value=value):
                self.assertCountEqual(
                    [
                        u.username
                        for u in f.filter_queryset(User.objects.all(), value)
                    ],
                    expected,
                )


class FilterToggleTests(TestCase):
    """Tests for :py:class:`FilterToggle`."""

    @classmethod
    def setUpTestData(cls) -> None:
        """Initialize class data."""
        super().setUpTestData()
        cls.playground.create_user("foo")
        cls.playground.create_user("foobar")
        cls.playground.create_user("bar")

    def test_metadata_default(self) -> None:
        """Test defaults for field metadata."""
        f = FilterToggle("User", q=Q(username__contains="foo"))
        f.name = "username"
        self.assertEqual(f.label, "User")
        self.assertIsInstance(f.q, Q)
        self.assertIsNone(f.icon)
        self.assertIs(f.menu_entry, MenuEntryToggle)
        self.assertIs(f.main_entry, MainEntryToggle)

    def test_metadata_full(self) -> None:
        """Test field metadata."""
        f = FilterToggle(
            "User",
            q=Q(username__contains="foo"),
            icon="icon",
        )
        self.assertEqual(f.label, "User")
        self.assertIsInstance(f.q, Q)
        self.assertEqual(f.icon, "icon")
        self.assertIs(f.menu_entry, MenuEntryToggle)
        self.assertIs(f.main_entry, MainEntryToggle)

    def test_filter_queryset(self) -> None:
        """Test filter_queryset."""
        f = FilterToggle("User", q=Q(username__contains="foo"))
        f.name = "username"
        self.assertCountEqual(
            [u.username for u in f.filter_queryset(User.objects.all(), True)],
            ["foo", "foobar"],
        )

    def test_filter_queryset_inactive(self) -> None:
        """Test filter_queryset with an inactive field."""
        f = FilterToggle("User", q=Q(username__contains="foo"))
        f.name = "username"
        self.assertCountEqual(
            [u.username for u in f.filter_queryset(User.objects.all(), None)],
            ["_system", "foo", "foobar", "bar"],
        )


class FilterSelectOneTests(TestCase):
    """Tests for :py:class:`FilterSelectOne`."""

    def test_metadata_default(self) -> None:
        """Test defaults for field metadata."""
        f = FilterSelectOne(
            "User",
        )
        self.assertEqual(f.label, "User")
        self.assertEqual(f.options, ())
        self.assertIsNone(f.icon)
        self.assertIs(f.menu_entry, MenuEntry)
        self.assertIs(f.main_entry, MainEntrySelectOne)

    def test_metadata_full(self) -> None:
        """Test field metadata."""
        option = SelectOption("test", "Test", Q(username="test"), "icon")
        f = FilterSelectOne(
            "User",
            icon="icon",
            options=[option],
        )
        self.assertEqual(f.label, "User")
        self.assertEqual(f.icon, "icon")
        self.assertEqual(f.options, [option])

    def test_filter_queryset(self) -> None:
        """Test filter_queryset."""
        f = FilterSelectOne("User")
        with self.assertRaisesRegex(
            ImproperlyConfigured,
            r"FilterSelectOne should have bound_class"
            r" set to a BoundFilterSelect",
        ):
            f.filter_queryset(User.objects.all(), "test")


class FilterFieldTests(TestCase):
    """Tests for :py:class:`FilterField`."""

    @classmethod
    def setUpTestData(cls) -> None:
        """Initialize class data."""
        super().setUpTestData()
        cls.playground.create_user("foo")
        cls.playground.create_user("foobar")
        cls.playground.create_user("bar")

    def test_metadata_default(self) -> None:
        """Test defaults for field metadata."""
        f = FilterField()
        f.name = "username"
        self.assertEqual(f.label, "")
        self.assertEqual(f.q_field, "username__contains")
        self.assertIsNone(f.icon)
        self.assertIs(f.menu_entry, MenuEntry)
        self.assertIs(f.main_entry, MainEntryField)

    def test_metadata_full(self) -> None:
        """Test field metadata."""
        f = FilterField("Test", "testicon", q_field="test")
        f.name = "username"
        self.assertEqual(f.label, "Test")
        self.assertEqual(f.q_field, "test")
        self.assertEqual(f.icon, "testicon")

    def test_filter_queryset(self) -> None:
        """Test filter_queryset."""
        f = FilterField()
        f.name = "username"
        self.assertCountEqual(
            [u.username for u in f.filter_queryset(User.objects.all(), "bar")],
            ["foobar", "bar"],
        )


class BoundFilterTests(TableTestCase):
    """Tests for :py:class:`BoundFilter`."""

    @classmethod
    def setUpTestData(cls) -> None:
        """Initialize class data."""
        super().setUpTestData()
        cls.playground.create_user("foo")
        cls.playground.create_user("foobar")
        cls.playground.create_user("bar")

    def get_table_class(self) -> type[Table[User]]:
        class _Table(Table[User]):
            user = Column("User", ordering="username")
            filter_user1 = FilterText("Username", q_field="username__contains")
            filter_user2 = FilterSelectOne("User")
            default_order = "user"

        return _Table

    def test_members(self) -> None:
        """Test object members."""
        table = self._table()
        f = table.filters["user1"]
        self.assertIs(f.filter, table.__class__.filter_definitions["user1"])
        self.assertIs(f.filters, table.filters)
        self.assertEqual(f.filter_prefix, "filter-user1")
        self.assertEqual(f.value, "")
        self.assertFalse(f.active)
        self.assertEqual(f.qs_remove, "")
        self.assertEqual(f.base_filter_args, [])
        self.assertEqual(f.name, "user1")
        self.assertEqual(f.label, "Username")
        self.assertIsNone(f.icon)
        self.assertEqual(f.qs_with_value(42), "filter-user1=42")
        self.assertIs(f.menu_entry.filter, f)
        self.assertIs(f.main_entry.filter, f)

    def test_table_prefix(self) -> None:
        """Ensure BoundFilter honors table_prefix."""
        table = self._table(prefix="test")
        f = table.filters["user1"]
        self.assertEqual(f.filter_prefix, "test-filter-user1")

    def test_preserve_order(self) -> None:
        """Ensure BoundFilter query string handling preserves ordering."""
        table = self._table(qstring={"order": "user"})
        f = table.filters["user1"]
        self.assertEqual(f.qs_remove, "order=user")
        self.assertEqual(f.base_filter_args, [("order", "user")])

    def test_preserve_other_filters(self) -> None:
        """Ensure BoundFilter query string handling preserves other filterse."""
        table = self._table(
            qstring={
                "order": "user",
                "filter-user1": "foo",
                "filter-user2": "bar",
            }
        )
        f = table.filters["user1"]
        self.assertEqual(f.qs_remove, "order=user&filter-user2=bar")
        self.assertEqual(
            f.base_filter_args, [("order", "user"), ("filter-user2", "bar")]
        )

    def test_value_formatted(self) -> None:
        table = self._table()
        f = table.filters["user1"]
        for value, expected in (
            (42, "42"),
            ("foo", "foo"),
            (None, ""),
            (["foo", "bar"], "foo, bar"),
        ):
            with self.subTest(value=value):
                f.value = value
                self.assertEqual(f._format_value(), expected)

    def test_value_for_querystring(self) -> None:
        table = self._table()
        f = table.filters["user1"]
        for value, expected in (
            (True, ["1"]),
            (None, []),
            (False, []),
            ("foo", ["foo"]),
            ([1, 2, 3], ["1", "2", "3"]),
            (42, ["42"]),
        ):
            with self.subTest(value=value):
                self.assertEqual(f.value_for_querystring(value), expected)

    def test_qs_with_value(self) -> None:
        table = self._table(
            qstring={
                "order": "user",
                "filter-user1": "foo",
                "filter-user2": "bar",
            }
        )
        f = table.filters["user1"]
        self.assertEqual(
            set(f.qs_with_value([1, 2]).split("&")),
            {
                "order=user",
                "filter-user1=1",
                "filter-user1=2",
                "filter-user2=bar",
            },
        )

    def test_qs_with_value_empty_value(self) -> None:
        """Calling qs_with_value with an empty value gives qs_remove."""
        table = self._table(
            qstring={
                "order": "user",
                "filter-user1": "foo",
                "filter-user2": "bar",
            }
        )
        f = table.filters["user1"]
        self.assertEqual(
            set(f.qs_with_value(None).split("&")),
            {
                "order=user",
                "filter-user2=bar",
            },
        )

    def test_add_option(self) -> None:
        """BoundFilter.add_option refuses to work."""
        table = self._table(
            qstring={
                "order": "user",
                "filter-user1": "foo",
                "filter-user2": "bar",
            }
        )
        f = table.filters["user1"]
        with self.assertRaisesRegex(
            ImproperlyConfigured,
            "add_option called on a field that does not support options.",
        ):
            f.add_option("test", "Test", Q(username="test"))

    def test_filter_queryset(self) -> None:
        """Test :py:func:`filter_queryset`."""
        table = self._table(qstring={"filter-user1": "foo"})
        f = table.filters["user1"]
        self.assertCountEqual(
            [u.username for u in f.filter_queryset(User.objects.all())],
            ["foo", "foobar"],
        )


class BoundFilterSelectOneTests(TableTestCase):
    """Tests for :py:class:`BoundFilterSelectOne`."""

    @classmethod
    def setUpTestData(cls) -> None:
        """Initialize class data."""
        super().setUpTestData()
        cls.playground.create_user("foo")
        cls.playground.create_user("foobar")
        cls.playground.create_user("bar")

    def get_table_class(self) -> type[Table[User]]:
        class _Table(Table[User]):
            user = Column("User", ordering="username")
            filter_user1 = FilterText("Username", q_field="username")
            filter_user2 = FilterSelectOne("User")
            filter_user3 = FilterSelectOne(
                "User",
                options=[
                    SelectOption("bar", "Bar", Q(username="bar")),
                ],
            )
            default_order = "user"

            def init(self) -> None:
                super().init()
                self.filters["user2"].add_option(
                    "foo", "Foo", Q(username="foo")
                )
                self.filters["user2"].add_option(
                    "bar", "Bar", Q(username="bar")
                )

        return _Table

    def test_members(self) -> None:
        """Test object members."""
        table = self._table()
        f = table.filters["user2"]
        self.assertIsInstance(f, BoundFilterSelectOne)
        assert isinstance(f, BoundFilterSelectOne)
        self.assertIs(f.filter, table.__class__.filter_definitions["user2"])
        self.assertIs(f.filters, table.filters)
        self.assertEqual(f.filter_prefix, "filter-user2")
        self.assertEqual(f.value, "")
        self.assertFalse(f.active)
        self.assertEqual(f.qs_remove, "")
        self.assertEqual(f.base_filter_args, [])
        self.assertEqual(f.name, "user2")
        self.assertEqual(f.label, "User")
        self.assertIsNone(f.icon)
        self.assertEqual(list(f.options.keys()), ["foo", "bar"])
        self.assertEqual(f.qs_with_value(42), "filter-user2=42")
        self.assertIs(f.menu_entry.filter, f)
        self.assertIs(f.main_entry.filter, f)

    def test_declarative_options(self) -> None:
        """Test setting filter options declaratively."""
        table = self._table()
        f = table.filters["user3"]
        assert isinstance(f, BoundFilterSelectOne)
        self.assertEqual(list(f.options.keys()), ["bar"])

    def test_filter_queryset(self) -> None:
        """Test :py:func:`filter_queryset`."""
        table = self._table(qstring={"filter-user2": "foo"})
        f = table.filters["user2"]
        self.assertCountEqual(
            [u.username for u in f.filter_queryset(User.objects.all())],
            ["foo"],
        )

    def test_filter_queryset_invalid_value(self) -> None:
        """Test :py:func:`filter_queryset`."""
        table = self._table(qstring={"filter_user2": "does-not-exist"})
        f = table.filters["user2"]
        self.assertCountEqual(
            {u.username for u in f.filter_queryset(User.objects.all())},
            {"_system", "foo", "foobar", "bar"},
        )


class BoundFilterFieldTests(TableTestCase):
    """Tests for :py:class:`BoundFilterField`."""

    @classmethod
    def setUpTestData(cls) -> None:
        """Initialize class data."""
        super().setUpTestData()
        cls.playground.create_user("foo")
        cls.playground.create_user("foobar")
        cls.playground.create_user("bar")

    def get_table_class(self) -> type[Table[User]]:
        class _Table(Table[User]):
            user = Column("User", ordering="username")
            filter_username = FilterField("Username")
            default_order = "user"

        return _Table

    def test_members(self) -> None:
        """Test object members."""
        table = self._table()
        f = table.filters["username"]
        self.assertIsInstance(f, BoundFilterField)
        assert isinstance(f, BoundFilterField)
        self.assertIs(f.filter, table.__class__.filter_definitions["username"])
        self.assertIs(f.filters, table.filters)
        self.assertEqual(f.filter_prefix, "filter-username")
        self.assertIsNone(f.value)
        self.assertFalse(f.active)
        self.assertEqual(f.qs_remove, "")
        self.assertEqual(f.base_filter_args, [])
        self.assertEqual(f.name, "username")
        self.assertEqual(f.label, "Username")
        self.assertIsNone(f.icon)
        self.assertEqual(f.qs_with_value(42), "filter-username=42")
        self.assertIs(f.menu_entry.filter, f)
        self.assertIs(f.main_entry.filter, f)

    def test_value_formatted(self) -> None:
        table = self._table()
        f = table.filters["username"]
        for value, expected in (
            (None, ""),
            ("foo", "foo"),
            ("invalid", "invalid"),
        ):
            with self.subTest(value=value):
                f.value = value
                self.assertEqual(f._format_value(), expected)

    def test_invalid_field_name_with_auto_form(self) -> None:
        """Test object members."""

        class _Table(Table[User]):
            user = Column("User", ordering="username")
            filter_mischief = FilterField()
            default_order = "user"

        table = self._table(table_class=_Table)
        f = table.filters["mischief"]
        with self.assertRaisesRegex(
            FieldError,
            re.escape("Unknown field(s) (mischief) specified for User"),
        ):
            f.value

    def test_invalid_field_name_with_custom_form(self) -> None:
        """Test object members."""

        class _Table(Table[User]):
            user = Column("User", ordering="username")
            filter_mischief = FilterField()
            filter_form_class = modelform_factory(User, fields="__all__")
            default_order = "user"

        table = self._table(table_class=_Table)
        f = table.filters["mischief"]
        with self.assertRaisesRegex(
            ImproperlyConfigured,
            r"'mischief' not found in table filter form UserForm."
            r" Available fields: \[.+\]",
        ):
            # Trigger instantiating the form field
            f.label

    def test_filter_queryset(self) -> None:
        """Test :py:func:`filter_queryset`."""
        table = self._table(qstring={"filter-username": "foo"})
        f = table.filters["username"]
        self.assertEqual(f.value, "foo")
        self.assertTrue(f.active)
        self.assertCountEqual(
            [u.username for u in f.filter_queryset(User.objects.all())],
            ["foo", "foobar"],
        )


class BoundFilterSelectFieldTests(TableTestCase):
    """Tests for :py:class:`BoundFilterField` using a select form field."""

    def get_table_class(self) -> type[Table[User]]:
        class _Form(Form):
            user = ChoiceField(
                choices=(
                    ("foo", "Foo"),
                    ("bar", "Bar"),
                    ("Multi", [("test1", "Test1"), ("test2", "Test2")]),
                )
            )

        class _Table(Table[User]):
            user = Column("User", ordering="username")
            filter_user = FilterField("Username")
            default_order = "user"
            filter_form_class = _Form

        return _Table

    def test_value_formatted(self) -> None:
        table = self._table()
        f = table.filters["user"]
        for value, expected in (
            (None, ""),
            ("foo", "Foo"),
            ("bar", "Bar"),
            ("test1", "Test1"),
            ("test2", "Test2"),
            (["foo", "test1"], "Foo, Test1"),
            ("invalid", "invalid"),
        ):
            with self.subTest(value=value):
                f.value = value
                self.assertEqual(f._format_value(), expected)


class FiltersTests(TableTestCase):
    """Tests for :py:class:`Filters`."""

    @classmethod
    def setUpTestData(cls) -> None:
        """Initialize class data."""
        super().setUpTestData()
        cls.playground.create_user("foo")
        cls.playground.create_user("foobar")
        cls.playground.create_user("bar")

    def get_table_class(self) -> type[Table[User]]:
        class _Table(Table[User]):
            user = Column("User", ordering="username")
            filter_user1 = FilterText("Username", q_field="username__contains")
            filter_user2 = FilterSelectOne("User")
            default_order = "user"

        return _Table

    def test_filters_fields(self) -> None:
        """Test Filters fields instantiated from Table."""
        table = self._table()
        filters = table.filters
        self.assertIs(filters.table, table)
        self.assertEqual(filters.filter_prefix, "filter")
        self.assertEqual(list(filters.options.keys()), ["user1", "user2"])
        self.assertEqual(filters.qs_clear_filters, "")
        self.assertEqual(filters.non_filter_args, [])
        self.assertTrue(filters)
        self.assertEqual(
            [f.name for f in filters.available], ["user1", "user2"]
        )
        self.assertEqual([f.name for f in filters.active], [])
        self.assertEqual(filters["user1"].name, "user1")
        self.assertEqual(filters["user2"].name, "user2")
        self.assertEqual(filters.get_context_data(), {"filters": filters})

    def test_qs_clear_filters(self) -> None:
        """qs_clear_filters preserves ordering elements."""
        table = self._table(qstring={"order": "user", "filter-user": "test"})
        filters = table.filters
        self.assertEqual(filters.qs_clear_filters, "order=user")
        self.assertEqual(filters.non_filter_args, [("order", "user")])

    def test_qs_clear_filters_table_prefix(self) -> None:
        """qs_clear_filters preserves ordering elements."""
        table = self._table(
            prefix="test",
            qstring={"test-order": "user", "test-filter-user": "test"},
        )
        filters = table.filters
        self.assertEqual(filters.filter_prefix, "test-filter")
        self.assertEqual(filters.qs_clear_filters, "test-order=user")
        self.assertEqual(filters.non_filter_args, [("test-order", "user")])

    def test_form_default(self) -> None:
        """Test autogenerating a filter form."""

        class _Table(Table[User]):
            user = Column("User", ordering="username")
            filter_username = FilterField()
            default_order = "user"

        table = self._table(table_class=_Table)
        form = table.filters.form
        self.assertCountEqual(
            form.fields.keys(),
            ["username"],
        )

    def test_form_default_skip_nonform_fields(self) -> None:
        """Skip non-form fields when autogenerating a filter form."""

        class _Table(Table[User]):
            user = Column("User", ordering="username")
            filter_username = FilterField()
            filter_email = FilterText("Email")
            default_order = "user"

        table = self._table(table_class=_Table)
        form = table.filters.form
        self.assertCountEqual(
            form.fields.keys(),
            ["username"],
        )

    def test_form_custom(self) -> None:
        """Test autogenerating a filter form."""
        form_class = modelform_factory(User, fields=["email", "username"])

        class _Table(Table[User]):
            user = Column("User", ordering="username")
            default_order = "user"
            filter_form_class = form_class

        table = self._table(table_class=_Table)
        form = table.filters.form
        self.assertCountEqual(
            form.fields.keys(),
            [
                "email",
                "username",
            ],
        )

    def test_filter_queryset(self) -> None:
        """Test :py:func:`filter_queryset`."""
        table = self._table(qstring={"filter-user1": "foo"})
        self.assertCountEqual(
            [
                u.username
                for u in table.filters.filter_queryset(User.objects.all())
            ],
            ["foo", "foobar"],
        )

    def test_render(self) -> None:
        """Test :py:func:`render`."""
        table = self._table()
        tree = self.assertHTMLValid(
            self.render_widget(table.filters), dump_on_error=True
        )
        div = self.assertHasElement(tree, "body/div")

        # Filter menu
        filters_opener = self.assertHasElement(
            div, "//div[@id='table-filters-menu']"
        )
        parent = filters_opener.getparent()
        assert parent is not None
        filters_menu = parent.ul[0]
        self.assertEqual(
            [self.get_node_text_normalized(li) for li in filters_menu.li],
            ["Add filter on field…", "Username", "User"],
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Views for the server application: collections."""

import heapq
import logging
import re
from functools import cached_property, total_ordering
from typing import Any, override

from django.db import IntegrityError
from django.shortcuts import get_object_or_404
from rest_framework import mixins, status
from rest_framework.decorators import action
from rest_framework.parsers import JSONParser
from rest_framework.request import Request
from rest_framework.response import Response

from debusine.artifacts.models import (
    BareDataCategory,
    CollectionCategory,
    DebusineTaskConfiguration,
    SINGLETON_COLLECTION_CATEGORIES,
)
from debusine.db.context import context
from debusine.db.models import Collection, CollectionItem
from debusine.db.models.collections import CollectionQuerySet
from debusine.server.collections import DebusineTaskConfigurationManager
from debusine.server.exceptions import DebusineAPIException
from debusine.server.serializers import (
    BaseSerializerBase,
    CollectionSerializer,
    TaskConfigurationCollectionContents,
    TaskConfigurationCollectionContentsSerializer,
    TaskConfigurationCollectionUpdateResultsSerializer,
)
from debusine.server.views.base import BaseAPIView, GenericViewSetBase
from debusine.server.views.rest import IsTokenUserAuthenticated

logger = logging.getLogger(__name__)


@total_ordering
class ComparableItem:
    """Sortable container for DebusineTaskConfiguration entries."""

    def __init__(
        self, item: DebusineTaskConfiguration, name: str | None = None
    ) -> None:
        """Store an item together with its database name."""
        self.item = item
        # Store the database name together with the item, for those items that
        # are in the database
        self.name = name

    @cached_property
    def _sort_key(self) -> tuple[Any, ...]:
        item = self.item
        return (
            item.task_type or "",
            item.task_name or "",
            item.subject or "",
            item.context or "",
            item.template or "",
            item.use_templates or [],
            item.delete_values or [],
            sorted(item.default_values.items()) if item.default_values else [],
            (
                sorted(item.override_values.items())
                if item.override_values
                else []
            ),
            item.lock_values or [],
            item.comment or "",
        )

    def __eq__(self, other: Any) -> bool:
        """Delegate equality to contained items."""
        return bool(self.item == other.item)

    def __lt__(self, other: Any) -> bool:
        """
        Provide ordering.

        Ordering is arbitrary, and provided to allow to use
        DebusineTaskConfiguration items in heapq or similar structures.
        """
        if not isinstance(other, ComparableItem):
            return NotImplemented
        return bool(self._sort_key < other._sort_key)


class TaskConfigurationCollectionView(BaseAPIView):
    """View used to fetch and update a task configuration collection."""

    permission_classes = [IsTokenUserAuthenticated]
    parser_classes = [JSONParser]

    def _get_collection(self, workspace: str, name: str) -> Collection:
        """Lookup the task configuration collection."""
        self.set_current_workspace(workspace)
        try:
            collection = context.require_workspace().get_collection(
                user=context.require_user(),
                category=CollectionCategory.TASK_CONFIGURATION,
                name=name,
            )
        except Collection.DoesNotExist:
            raise DebusineAPIException(
                title="Collection not found",
                detail=(
                    f"Task configuration collection {name!r} does not exist"
                    f" in workspace {workspace!r}"
                ),
                status_code=status.HTTP_404_NOT_FOUND,
            )

        self.enforce(collection.can_display)
        return collection

    def get(
        self,
        request: Request,  # noqa: ARG002, U100
        workspace: str,
        name: str,
    ) -> Response:
        """Retrieve the contents of the collection."""
        collection = self._get_collection(workspace, name)

        items: list[DebusineTaskConfiguration] = []
        for child in collection.child_items.active().filter(
            child_type=CollectionItem.Types.BARE
        ):
            items.append(DebusineTaskConfiguration(**child.data))

        contents = TaskConfigurationCollectionContents(
            collection=collection, items=items
        )
        return Response(
            TaskConfigurationCollectionContentsSerializer(contents).data,
            status=status.HTTP_200_OK,
        )

    @staticmethod
    def _diff_items(
        collection: Collection, items: list[DebusineTaskConfiguration]
    ) -> tuple[list[str], list[DebusineTaskConfiguration], int]:
        """
        Compare the collection contents against a new list of items.

        :returns: a tuple (removed, added), where removed lists the item IDs to
                  remove, and added the DebusineTaskConfiguration items to add.
        """
        # Load the existing items
        old: list[ComparableItem] = []
        for child in collection.child_items.active().filter(
            child_type=CollectionItem.Types.BARE
        ):
            old.append(
                ComparableItem(
                    item=DebusineTaskConfiguration(**child.data),
                    name=child.name,
                )
            )

        # Wrap the new items for comparison
        new: list[ComparableItem] = [ComparableItem(item=i) for i in items]

        # DebusineTaskConfiguration objects are not hashable, so we cannot use
        # sets. We work around that by making them comparable, and using heapq
        heapq.heapify(old)
        heapq.heapify(new)

        removed: list[str] = []
        added: list[DebusineTaskConfiguration] = []
        unchanged: int = 0
        while old and new:
            if old[0] == new[0]:
                heapq.heappop(old)
                heapq.heappop(new)
                unchanged += 1
            elif old[0] < new[0]:
                assert old[0].name is not None
                removed.append(old[0].name)
                heapq.heappop(old)
            else:
                added.append(new[0].item)
                heapq.heappop(new)

        for i in old:
            assert i.name is not None
            removed.append(i.name)
        added += [i.item for i in new]

        return removed, added, unchanged

    def _validate_git_commit(self, value: Any, name: str) -> str:
        if not isinstance(value, str):
            raise DebusineAPIException(
                title="Invalid git commit",
                detail=f"{name} is not a string",
                status_code=status.HTTP_400_BAD_REQUEST,
            )
        if not re.match(r"^[0-9a-f]{40,64}$", value):
            raise DebusineAPIException(
                title="Invalid git commit",
                detail=f"{name} is malformed",
                status_code=status.HTTP_400_BAD_REQUEST,
            )
        return value

    def post(self, request: Request, workspace: str, name: str) -> Response:
        """Update the contents of the collection."""
        collection = self._get_collection(workspace, name)
        self.enforce(collection.workspace.can_edit_task_configuration)
        dry_run = request.data.pop("dry_run", False)
        serializer = TaskConfigurationCollectionContentsSerializer(
            data=request.data
        )
        serializer.is_valid(raise_exception=True)
        submitted_collection = serializer.validated_data["collection"]

        if submitted_collection["id"] != collection.pk:
            raise DebusineAPIException(
                title="Posted collection mismatch",
                detail=(
                    f"Data posted for collection {submitted_collection['id']}"
                    f" to update collection {collection.pk}"
                ),
                status_code=status.HTTP_400_BAD_REQUEST,
            )

        if submitted_collection["name"] != collection.name:
            raise DebusineAPIException(
                title="Posted collection mismatch",
                detail=(
                    "Data posted for collection"
                    f" {submitted_collection['name']!r}"
                    f" to update collection {collection.name!r}"
                ),
                status_code=status.HTTP_400_BAD_REQUEST,
            )

        # Update git commit in collection data
        if (
            git_commit := submitted_collection["data"].get("git_commit")
        ) is not None:
            git_commit = self._validate_git_commit(git_commit, "git_commit")
            collection.data["git_commit"] = git_commit
        else:
            collection.data.pop("git_commit", None)
        collection.save()

        # We do not update existing items, as that would rewrite collection
        # history. If an item has been changed, we mark the old version as
        # deleted and create a new version
        removed, added, unchanged = self._diff_items(
            collection, serializer.validated_data["items"]
        )

        # Update collection
        manager = DebusineTaskConfigurationManager(collection=collection)
        user = context.require_user()
        assert user.is_authenticated
        if not dry_run:
            for name in removed:
                manager.remove_items_by_name(
                    name=name,
                    child_types=[CollectionItem.Types.BARE],
                    user=user,
                )

        # Count updates as the number of added names that are also in removed
        updated: int = 0

        for item in added:
            if not dry_run:
                name = manager.add_bare_data(
                    BareDataCategory.TASK_CONFIGURATION, user=user, data=item
                ).name
            else:
                name = item.name()
            if name in removed:
                updated += 1

        return Response(
            TaskConfigurationCollectionUpdateResultsSerializer(
                {
                    "added": len(added) - updated,
                    "removed": len(removed) - updated,
                    "updated": updated,
                    "unchanged": unchanged,
                }
            ).data,
            status=status.HTTP_200_OK,
        )


class CollectionViewSet(
    BaseAPIView,
    mixins.CreateModelMixin,
    mixins.RetrieveModelMixin,
    mixins.ListModelMixin,
    mixins.UpdateModelMixin,
    mixins.DestroyModelMixin,
    GenericViewSetBase[Collection],
):
    """Provide access to collections."""

    permission_classes = [IsTokenUserAuthenticated]
    parser_classes = [JSONParser]
    serializer_class = CollectionSerializer

    def initial(self, request: Request, *args: Any, **kwargs: Any) -> None:
        """Return the initial request object."""
        super().initial(request, *args, **kwargs)
        self.set_current_workspace(kwargs["workspace"])

    def get_queryset(self) -> CollectionQuerySet[Any]:
        """Filter accessible collections."""
        queryset = (
            Collection.objects.can_display(context.user)
            .filter(workspace=context.require_workspace())
            .order_by("category", "name")
        )
        if self.action == "list":
            queryset = queryset.exclude(
                category=CollectionCategory.WORKFLOW_INTERNAL
            )
        return queryset

    @override
    def get_serializer_context(self) -> dict[str, Any]:
        context = super().get_serializer_context()
        context['with_stats'] = self.request.GET.get("stats") == "1"
        return context

    @action(
        detail=False,
        url_path=r"(?P<category>[^/]+)/(?P<name>[^/]+)",
        url_name="lookup",
    )
    def lookup(
        self,
        request: Request,  # noqa: ARG002, U100
        workspace: str,  # noqa: ARG002, U100
        category: str,
        name: str,
    ) -> Response:
        """Look up a collection by name."""
        collection = get_object_or_404(
            self.get_queryset(), name=name, category=category
        )
        serializer = self.get_serializer(collection)
        return Response(serializer.data)

    def set_name(self, collection: Collection, name: str) -> bool:
        """
        Rename the collection.

        :return: True if the name has been changed
        """
        if name == collection.name:
            return False

        if (
            collection.category in SINGLETON_COLLECTION_CATEGORIES
            and name != "_"
        ):
            raise DebusineAPIException(
                title="Invalid name for singleton collection",
                detail=(
                    f"'{collection.category}' singleton collections"
                    " must be named '_'"
                ),
                status_code=status.HTTP_400_BAD_REQUEST,
            )

        if Collection.objects.filter(
            workspace=collection.workspace,
            category=collection.category,
            name=name,
        ).exists():
            raise DebusineAPIException(
                title="New name conflicts with an existing collection",
                detail=(
                    "A collection with category"
                    f" '{collection.category}' called {name!r}"
                    " already exists"
                ),
                status_code=status.HTTP_400_BAD_REQUEST,
            )

        collection.name = name
        return True

    def set_full_history_retention_period(
        self, collection: Collection, value: int | None
    ) -> bool:
        """
        Set the full history retention period.

        :return: True if the value has been changed
        """
        if value == collection.full_history_retention_period:
            return False
        collection.full_history_retention_period = value
        return True

    def set_metadata_only_retention_period(
        self, collection: Collection, value: int | None
    ) -> bool:
        """
        Set the full history retention period.

        :return: True if the value has been changed
        """
        if value == collection.metadata_only_retention_period:
            return False
        collection.metadata_only_retention_period = value
        return True

    def set_data(self, collection: Collection, value: dict[str, Any]) -> bool:
        """
        Set the collection data.

        :return: True if the value has been changed
        """
        if value == collection.data:
            return False
        # TODO: The key-value data should be validated against the
        # collection's pydantic data model, once we have it.
        # See https://salsa.debian.org/freexian-team/debusine/-/issues/923
        collection.data = value
        return True

    @override
    def perform_create(self, serializer: BaseSerializerBase[Any]) -> None:
        self.enforce(context.require_workspace().can_create_collections)
        try:
            serializer.save(workspace=context.require_workspace())
        except IntegrityError as e:
            detail = None
            message = str(e)
            if "db_collection_unique_name_category_workspace" in message:
                detail = "A collection with the same name already exists"
            raise DebusineAPIException(
                title="Cannot create collection", detail=detail
            )

    @override
    def perform_update(self, serializer: BaseSerializerBase[Any]) -> None:
        assert isinstance(serializer, CollectionSerializer)
        collection = serializer.instance
        assert isinstance(collection, Collection)
        self.enforce(collection.can_configure)

        changed: bool = False
        if "category" in serializer.validated_data:
            if serializer.validated_data["category"] != collection.category:
                raise DebusineAPIException(
                    title="The category of a collection cannot be changed",
                    status_code=status.HTTP_400_BAD_REQUEST,
                )
        if "name" in serializer.validated_data:
            changed = (
                self.set_name(collection, serializer.validated_data["name"])
                or changed
            )
        if "full_history_retention_period" in serializer.validated_data:
            changed = (
                self.set_full_history_retention_period(
                    collection,
                    serializer.validated_data["full_history_retention_period"],
                )
                or changed
            )
        if "metadata_only_retention_period" in serializer.validated_data:
            changed = (
                self.set_metadata_only_retention_period(
                    collection,
                    serializer.validated_data["metadata_only_retention_period"],
                )
                or changed
            )
        if "data" in serializer.validated_data:
            changed = (
                self.set_data(collection, serializer.validated_data["data"])
                or changed
            )

        if changed:
            collection.save()

    @override
    def perform_destroy(self, instance: Collection) -> None:
        self.enforce(instance.can_delete)
        instance.delete_collection()
